#
#  Copyright 2024 The InfiniFlow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import copy
import logging
import math
import os
import re
import string

import datrie
import nltk
from hanziconv import HanziConv
from nltk import PorterStemmer, WordNetLemmatizer, word_tokenize
from nltk.data import find


# import warnings

# # 忽略无效转义序列的警告
# warnings.filterwarnings(
#     "ignore", message=".*invalid escape sequence.*", category=DeprecationWarning
# )
#
# # 忽略pkg_resources弃用警告
# warnings.filterwarnings(
#     "ignore", message=".*pkg_resources is deprecated.*", category=DeprecationWarning
# )


def ensure_nltk_resource(resource_name):
    try:
        find(resource_name)
    except LookupError:
        logging.info(f"Resource '{resource_name}' not found. Downloading...")
        nltk.download(resource_name.split("/")[-1])


ensure_nltk_resource("tokenizers/punkt_tab")
ensure_nltk_resource("corpora/wordnet")
ensure_nltk_resource("corpora/omw-1.4")

_curr_dir = os.path.dirname(os.path.abspath(__file__))


class Tokenizer:
    def key_(self, line):
        return str(line.lower().encode("utf-8"))[2:-1]

    def rkey_(self, line):
        return str(("DD" + (line[::-1].lower())).encode("utf-8"))[2:-1]

    def add_tok(self, token, freq, tag):
        k = self.key_(token)
        F = int(math.log(float(freq) / self.DENOMINATOR) + 0.5)
        if k not in self.trie_ or self.trie_[k][0] < F:
            self.trie_[self.key_(token)] = (F, tag)
        self.trie_[self.rkey_(token)] = 1

    def _load_dict(self, fnm):
        logging.info(f"[HUQIE]:Build trie from {fnm}")
        try:
            of = open(fnm, "r", encoding="utf-8")
            while True:
                line = of.readline()
                if not line:
                    break
                line = re.sub(r"[\r\n]+", "", line)
                line = re.split(r"[ \t]", line)
                self.add_tok(line[0], line[1], line[2])

            # dict_file_cache = fnm + ".trie"
            # logging.info(f"[HUQIE]:Build trie cache to {dict_file_cache}")
            # self.trie_.save(dict_file_cache)
            of.close()
        except Exception:
            logging.exception(f"[HUQIE]:Build trie {fnm} failed")

    def __init__(self, debug=False):
        self.DEBUG = debug
        self.DENOMINATOR = 1000000
        self.DIR_ = _curr_dir + "/dictionary"

        self.stemmer = PorterStemmer()
        self.lemmatizer = WordNetLemmatizer()

        self.SPLIT_CHAR = r"([ ,\.<>/?;:'\[\]\\`!@#$%^&*\(\)\{\}\|_+=《》，。？、；‘’：“”【】~！￥%……（）——-]+|[a-zA-Z0-9,\.-]+)"

        self.trie_ = datrie.Trie(string.printable)
        self._load_dict(self.DIR_ + "/huqie.txt")

    def load_user_dict(self, fnm):
        try:
            self.trie_ = datrie.Trie.load(fnm + ".trie")
            return
        except Exception:
            self.trie_ = datrie.Trie(string.printable)
        self._load_dict(fnm)

    def add_user_dict(self, fnm):
        self._load_dict(fnm)

    def _strQ2B(self, ustring):
        """Convert full-width characters to half-width characters"""
        rstring = ""
        for uchar in ustring:
            inside_code = ord(uchar)
            if inside_code == 0x3000:
                inside_code = 0x0020
            else:
                inside_code -= 0xFEE0
            if (
                inside_code < 0x0020 or inside_code > 0x7E
            ):  # After the conversion, if it's not a half-width character, return the original character.
                rstring += uchar
            else:
                rstring += chr(inside_code)
        return rstring

    strQ2B = _strQ2B

    def _tradi2simp(self, line):
        return HanziConv.toSimplified(line)

    tradi2simp = _tradi2simp

    def dfs_(self, chars, s, preTks, tkslist, _depth=0, _memo=None):
        if _memo is None:
            _memo = {}
        MAX_DEPTH = 10
        if _depth > MAX_DEPTH:
            if s < len(chars):
                copy_pretks = copy.deepcopy(preTks)
                remaining = "".join(chars[s:])
                copy_pretks.append((remaining, (-12, "")))
                tkslist.append(copy_pretks)
            return s

        state_key = (s, tuple(tk[0] for tk in preTks)) if preTks else (s, None)
        if state_key in _memo:
            return _memo[state_key]

        res = s
        if s >= len(chars):
            tkslist.append(preTks)
            _memo[state_key] = s
            return s
        if s < len(chars) - 4:
            is_repetitive = True
            char_to_check = chars[s]
            for i in range(1, 5):
                if s + i >= len(chars) or chars[s + i] != char_to_check:
                    is_repetitive = False
                    break
            if is_repetitive:
                end = s
                while end < len(chars) and chars[end] == char_to_check:
                    end += 1
                mid = s + min(10, end - s)
                t = "".join(chars[s:mid])
                k = self.key_(t)
                copy_pretks = copy.deepcopy(preTks)
                if k in self.trie_:
                    copy_pretks.append((t, self.trie_[k]))
                else:
                    copy_pretks.append((t, (-12, "")))
                next_res = self.dfs_(
                    chars, mid, copy_pretks, tkslist, _depth + 1, _memo
                )
                res = max(res, next_res)
                _memo[state_key] = res
                return res

        S = s + 1
        if s + 2 <= len(chars):
            t1 = "".join(chars[s : s + 1])
            t2 = "".join(chars[s : s + 2])
            if self.trie_.has_keys_with_prefix(
                self.key_(t1)
            ) and not self.trie_.has_keys_with_prefix(self.key_(t2)):
                S = s + 2
        if (
            len(preTks) > 2
            and len(preTks[-1][0]) == 1
            and len(preTks[-2][0]) == 1
            and len(preTks[-3][0]) == 1
        ):
            t1 = preTks[-1][0] + "".join(chars[s : s + 1])
            if self.trie_.has_keys_with_prefix(self.key_(t1)):
                S = s + 2

        for e in range(S, len(chars) + 1):
            t = "".join(chars[s:e])
            k = self.key_(t)
            if e > s + 1 and not self.trie_.has_keys_with_prefix(k):
                break
            if k in self.trie_:
                pretks = copy.deepcopy(preTks)
                pretks.append((t, self.trie_[k]))
                res = max(res, self.dfs_(chars, e, pretks, tkslist, _depth + 1, _memo))

        if res > s:
            _memo[state_key] = res
            return res

        t = "".join(chars[s : s + 1])
        k = self.key_(t)
        copy_pretks = copy.deepcopy(preTks)
        if k in self.trie_:
            copy_pretks.append((t, self.trie_[k]))
        else:
            copy_pretks.append((t, (-12, "")))
        result = self.dfs_(chars, s + 1, copy_pretks, tkslist, _depth + 1, _memo)
        _memo[state_key] = result
        return result

    def freq(self, tk):
        k = self.key_(tk)
        if k not in self.trie_:
            return 0
        return int(math.exp(self.trie_[k][0]) * self.DENOMINATOR + 0.5)

    def tag(self, tk):
        k = self.key_(tk)
        if k not in self.trie_:
            return ""
        return self.trie_[k][1]

    def score_(self, tfts):
        B = 30
        F, L, tks = 0, 0, []
        for tk, (freq, tag) in tfts:
            F += freq
            L += 0 if len(tk) < 2 else 1
            tks.append(tk)
        # F /= len(tks)
        L /= len(tks)
        logging.debug(
            "[SC] {} {} {} {} {}".format(tks, len(tks), L, F, B / len(tks) + L + F)
        )
        return tks, B / len(tks) + L + F

    def _sort_tokens(self, tkslist):
        res = []
        for tfts in tkslist:
            tks, s = self.score_(tfts)
            res.append((tks, s))
        return sorted(res, key=lambda x: x[1], reverse=True)

    def merge_(self, tks):
        # if split chars is part of token
        res = []
        tks = re.sub(r"[ ]+", " ", tks).split()
        s = 0
        while True:
            if s >= len(tks):
                break
            E = s + 1
            for e in range(s + 2, min(len(tks) + 2, s + 6)):
                tk = "".join(tks[s:e])
                if re.search(self.SPLIT_CHAR, tk) and self.freq(tk):
                    E = e
            res.append("".join(tks[s:E]))
            s = E

        return " ".join(res)

    def _max_forward(self, line):
        res = []
        s = 0
        while s < len(line):
            e = s + 1
            t = line[s:e]
            while e < len(line) and self.trie_.has_keys_with_prefix(self.key_(t)):
                e += 1
                t = line[s:e]

            while e - 1 > s and self.key_(t) not in self.trie_:
                e -= 1
                t = line[s:e]

            if self.key_(t) in self.trie_:
                res.append((t, self.trie_[self.key_(t)]))
            else:
                res.append((t, (0, "")))

            s = e

        return self.score_(res)

    def _max_backward(self, line):
        res = []
        s = len(line) - 1
        while s >= 0:
            e = s + 1
            t = line[s:e]
            while s > 0 and self.trie_.has_keys_with_prefix(self.rkey_(t)):
                s -= 1
                t = line[s:e]

            while s + 1 < e and self.key_(t) not in self.trie_:
                s += 1
                t = line[s:e]

            if self.key_(t) in self.trie_:
                res.append((t, self.trie_[self.key_(t)]))
            else:
                res.append((t, (0, "")))

            s -= 1

        return self.score_(res[::-1])

    def english_normalize_(self, tks):
        return [
            (
                self.stemmer.stem(self.lemmatizer.lemmatize(t))
                if re.match(r"[a-zA-Z_-]+$", t)
                else t
            )
            for t in tks
        ]

    def _split_by_lang(self, line):
        txt_lang_pairs = []
        arr = re.split(self.SPLIT_CHAR, line)
        for a in arr:
            if not a:
                continue
            s = 0
            e = s + 1
            zh = is_chinese(a[s])
            while e < len(a):
                _zh = is_chinese(a[e])
                if _zh == zh:
                    e += 1
                    continue
                txt_lang_pairs.append((a[s:e], zh))
                s = e
                e = s + 1
                zh = _zh
            if s >= len(a):
                continue
            txt_lang_pairs.append((a[s:e], zh))
        return txt_lang_pairs

    def tokenize(self, line):
        line = re.sub(r"\W+", " ", line)
        line = self._strQ2B(line).lower()
        line = self._tradi2simp(line)

        arr = self._split_by_lang(line)
        res = []
        for L, lang in arr:
            if not lang:

                res.extend(
                    [
                        self.stemmer.stem(self.lemmatizer.lemmatize(t))
                        for t in word_tokenize(L)
                    ]
                )
                continue
            if len(L) < 2 or re.match(r"[a-z\.-]+$", L) or re.match(r"[0-9\.-]+$", L):
                res.append(L)
                continue

            # use maxforward for the first time
            tks, s = self._max_forward(L)
            tks1, s1 = self._max_backward(L)
            if self.DEBUG:
                logging.debug("[FW] {} {}".format(tks, s))
                logging.debug("[BW] {} {}".format(tks1, s1))

            i, j, _i, _j = 0, 0, 0, 0
            same = 0
            while (
                i + same < len(tks1)
                and j + same < len(tks)
                and tks1[i + same] == tks[j + same]
            ):
                same += 1
            if same > 0:
                res.append(" ".join(tks[j : j + same]))
            _i = i + same
            _j = j + same
            j = _j + 1
            i = _i + 1

            while i < len(tks1) and j < len(tks):
                tk1, tk = "".join(tks1[_i:i]), "".join(tks[_j:j])
                if tk1 != tk:
                    if len(tk1) > len(tk):
                        j += 1
                    else:
                        i += 1
                    continue

                if tks1[i] != tks[j]:
                    i += 1
                    j += 1
                    continue
                # backward tokens from_i to i are different from forward tokens from _j to j.
                tkslist = []
                self.dfs_("".join(tks[_j:j]), 0, [], tkslist)
                res.append(" ".join(self._sort_tokens(tkslist)[0][0]))

                same = 1
                while (
                    i + same < len(tks1)
                    and j + same < len(tks)
                    and tks1[i + same] == tks[j + same]
                ):
                    same += 1
                res.append(" ".join(tks[j : j + same]))
                _i = i + same
                _j = j + same
                j = _j + 1
                i = _i + 1

            if _i < len(tks1):
                assert _j < len(tks)
                assert "".join(tks1[_i:]) == "".join(tks[_j:])
                tkslist = []
                self.dfs_("".join(tks[_j:]), 0, [], tkslist)
                res.append(" ".join(self._sort_tokens(tkslist)[0][0]))

        res = " ".join(res)
        logging.debug("[TKS] {}".format(self.merge_(res)))
        return self.merge_(res)

    def fine_grained_tokenize(self, tks):
        tks = tks.split()
        zh_num = len([1 for c in tks if c and is_chinese(c[0])])
        if zh_num < len(tks) * 0.2:
            res = []
            for tk in tks:
                res.extend(tk.split("/"))
            return " ".join(res)

        res = []
        for tk in tks:
            if len(tk) < 3 or re.match(r"[0-9,\.-]+$", tk):
                res.append(tk)
                continue
            tkslist = []
            if len(tk) > 10:
                tkslist.append(tk)
            else:
                self.dfs_(tk, 0, [], tkslist)
            if len(tkslist) < 2:
                res.append(tk)
                continue
            stk = self._sort_tokens(tkslist)[1][0]
            if len(stk) == len(tk):
                stk = tk
            else:
                if re.match(r"[a-z\.-]+$", tk):
                    for t in stk:
                        if len(t) < 3:
                            stk = tk
                            break
                    else:
                        stk = " ".join(stk)
                else:
                    stk = " ".join(stk)

            res.append(stk)

        return " ".join(self.english_normalize_(res))


def is_chinese(s):
    if s >= "\u4e00" and s <= "\u9fa5":
        return True
    else:
        return False


def is_number(s):
    if s >= "\u0030" and s <= "\u0039":
        return True
    else:
        return False


def is_alphabet(s):
    if ("\u0041" <= s <= "\u005a") or ("\u0061" <= s <= "\u007a"):
        return True
    else:
        return False


def naive_qie(txt):
    tks = []
    for t in txt.split():
        if tks and re.match(r".*[a-zA-Z]$", tks[-1]) and re.match(r".*[a-zA-Z]$", t):
            tks.append(" ")
        tks.append(t)
    return tks



from typing import List

import numpy as np


def mmr_reranking_optimized(
    query_embedding: List[float],
    search_results: List[List[float]],
    lambda_param: float = 0.5,
    top_k: int = 10,
) -> List[int]:
    """
    | 场景描述 | 推荐 `lambda_param` | 解释与示例 |
    | :--- | :--- | :--- |
    | **传统搜索引擎** | **0.6 ~ 0.8** | 用户希望最相关的结果排在前面。可以接受一定程度重复，但第一页结果必须高度相关。 |
    | **推荐系统 / 探索发现** | **0.3 ~ 0.5** | 目标是展示多样化的内容，避免给用户推荐相同的东西。例如：新闻推荐、商品推荐、内容发现平台。 |
    | **学术检索、法律案例检索** | **0.5 ~ 0.7** | 需要平衡：既要找到最相关的文献/案例，又要覆盖问题的不同方面或不同学派的观点。 |
    | **去除重复/近乎重复的文档** | **0.1 ~ 0.3** | 主要目的是过滤掉内容几乎相同的冗余文档。 |
    """
    if not search_results:
        return []

    n_docs = len(search_results)
    top_k = min(top_k, n_docs)

    doc_embeddings_np = np.array(search_results)
    query_embedding_np = np.array(query_embedding)

    query_norm = np.linalg.norm(query_embedding_np)
    doc_norms = np.linalg.norm(doc_embeddings_np, axis=1)
    relevance_scores = np.dot(doc_embeddings_np, query_embedding_np) / (
        doc_norms * query_norm + 1e-9
    )

    doc_doc_similarities = np.dot(doc_embeddings_np, doc_embeddings_np.T) / (
        np.outer(doc_norms, doc_norms) + 1e-9
    )

    selected_indices = []
    remaining_indices = list(range(n_docs))

    first_selected = np.argmax(relevance_scores)
    selected_indices.append(first_selected)
    remaining_indices.remove(first_selected)

    while len(selected_indices) < top_k and remaining_indices:
        rel_scores = relevance_scores[remaining_indices]
        max_sim_to_selected = np.max(
            doc_doc_similarities[np.ix_(remaining_indices, selected_indices)], axis=1
        )
        mmr_scores = (
            lambda_param * rel_scores - (1 - lambda_param) * max_sim_to_selected
        )

        best_idx_in_remaining = np.argmax(mmr_scores)
        best_idx = remaining_indices[best_idx_in_remaining]

        selected_indices.append(best_idx)
        del remaining_indices[best_idx_in_remaining]

    return selected_indices



from typing import Optional, Dict

from pydantic import BaseModel, Field


class FileParams(BaseModel):
    file_path: str
    content: str | bytes
    permissions: Optional[str] = Field(default="644")
    description: Optional[str] = Field(default="")

    @property
    def is_bytes(self):
        return isinstance(self.content, bytes)


class Resources(BaseModel):
    files: Optional[Dict[str, FileParams]] = {}

    def file_add(
        self,
        file_path: str,
        content: str | bytes,
        permissions: Optional[str] = "644",
        description: Optional[str] = "",
    ):
        self.files[file_path] = FileParams(
            file_path=file_path,
            content=content,
            permissions=permissions,
            description=description,
        )

    def file_str_replace(self, file_path: str, old_str: str, new_str: str):
        _file = self.files[file_path]
        if _file.is_bytes:
            raise ValueError("file is bytes, can not replace")

        if old_str in self.files[file_path].content:
            self.files[file_path].content = self.files[file_path].content.replace(
                old_str, new_str
            )
            return True
        else:
            return False

    def file_full_rewrite(
        self,
        file_path: str,
        content: str | bytes,
        permissions: Optional[str] = "644",
        description: Optional[str] = "",
    ):
        self.files[file_path] = FileParams(
            file_path=file_path,
            content=content,
            permissions=permissions,
            description=description,
        )

    def file_delete(self, file_path: str):
        if file_path in self.files[file_path]:
            del self.files[file_path]

    def file_exists(self, file_path: str):
        return file_path in self.files

    def read_file(self, file_path: str, start_line: int, end_line: int):
        # 通过路径判断文件是否可读,例如 .txt .csv等
        _file = self.files[file_path]
        if _file.is_bytes:
            raise ValueError("file is bytes, can not read")

        _lines = self.files[file_path].content.split("\n")
        return {
            "content": "\n".join(_lines[start_line:end_line]),
            "start_line": start_line,
            "end_line": end_line if end_line is not None else len(_lines),
            "total_lines": len(_lines),
        }

    def read_all_file(self, file_path: str):
        return self.files[file_path].content



"""
XML Tool Call Parser Module

This module provides a reliable XML tool call parsing system that supports
the Cursor-style format with structured function_calls blocks.
"""

import json
import logging
import re
from dataclasses import dataclass
from typing import List, Dict, Any, Optional, Tuple

logger = logging.getLogger(__name__)


@dataclass
class XMLToolCall:
    """Represents a parsed XML tool call."""

    function_name: str
    parameters: Dict[str, Any]
    raw_xml: str
    parsing_details: Dict[str, Any]


class XMLToolParser:
    """
    Parser for XML tool calls using the Cursor-style format:

    <function_calls>
    <invoke name="function_name">
    <parameter name="param_name">param_value</parameter>
    ...
    </invoke>
    </function_calls>
    """

    # Regex patterns for extracting XML blocks
    FUNCTION_CALLS_PATTERN = re.compile(
        r"<function_calls>(.*?)</function_calls>", re.DOTALL | re.IGNORECASE
    )

    INVOKE_PATTERN = re.compile(
        r'<invoke\s+name=["\']([^"\']+)["\']>(.*?)</invoke>', re.DOTALL | re.IGNORECASE
    )

    PARAMETER_PATTERN = re.compile(
        r'<parameter\s+name=["\']([^"\']+)["\']>(.*?)</parameter>',
        re.DOTALL | re.IGNORECASE,
    )

    def __init__(self, strict_mode: bool = False):
        """
        Initialize the XML tool parser.

        Args:
            strict_mode: If True, only accept the exact format. If False,
                        also try to parse legacy formats for backwards compatibility.
        """
        self.strict_mode = strict_mode

    def parse_content(self, content: str) -> List[XMLToolCall]:
        """
        Parse XML tool calls from content.

        Args:
            content: The text content potentially containing XML tool calls

        Returns:
            List of parsed XMLToolCall objects
        """
        tool_calls = []

        # First, try to find function_calls blocks
        function_calls_matches = self.FUNCTION_CALLS_PATTERN.findall(content)

        for fc_content in function_calls_matches:
            # Find all invoke blocks within this function_calls block
            invoke_matches = self.INVOKE_PATTERN.findall(fc_content)

            for function_name, invoke_content in invoke_matches:
                try:
                    tool_call = self._parse_invoke_block(
                        function_name, invoke_content, fc_content
                    )
                    if tool_call:
                        tool_calls.append(tool_call)
                except Exception as e:
                    logger.error(f"Error parsing invoke block for {function_name}: {e}")

        # If not in strict mode and no tool calls found, try legacy format
        if not self.strict_mode and not tool_calls:
            tool_calls.extend(self._parse_legacy_format(content))

        return tool_calls

    def _parse_invoke_block(
        self, function_name: str, invoke_content: str, full_block: str
    ) -> Optional[XMLToolCall]:
        """Parse a single invoke block into an XMLToolCall."""
        parameters = {}
        parsing_details = {
            "format": "v2",
            "function_name": function_name,
            "raw_parameters": {},
        }

        # Extract all parameters
        param_matches = self.PARAMETER_PATTERN.findall(invoke_content)

        for param_name, param_value in param_matches:
            # Clean up the parameter value
            param_value = param_value.strip()

            # Try to parse as JSON if it looks like JSON
            parsed_value = self._parse_parameter_value(param_value)

            parameters[param_name] = parsed_value
            parsing_details["raw_parameters"][param_name] = param_value

        # Extract the raw XML for this specific invoke
        invoke_pattern = re.compile(
            rf'<invoke\s+name=["\']{re.escape(function_name)}["\']>.*?</invoke>',
            re.DOTALL | re.IGNORECASE,
        )
        raw_xml_match = invoke_pattern.search(full_block)
        raw_xml = (
            raw_xml_match.group(0)
            if raw_xml_match
            else f'<invoke name="{function_name}">...</invoke>'
        )

        return XMLToolCall(
            function_name=function_name,
            parameters=parameters,
            raw_xml=raw_xml,
            parsing_details=parsing_details,
        )

    def _parse_parameter_value(self, value: str) -> Any:
        """
        Parse a parameter value, attempting to convert to appropriate type.

        Args:
            value: The string value to parse

        Returns:
            Parsed value (could be dict, list, bool, int, float, or str)
        """
        value = value.strip()

        # Try to parse as JSON first
        if value.startswith(("{", "[")):
            try:
                return json.loads(value)
            except json.JSONDecodeError:
                pass

        # Try to parse as boolean
        if value.lower() in ("true", "false"):
            return value.lower() == "true"

        # Try to parse as number
        try:
            if "." in value:
                return float(value)
            else:
                return int(value)
        except ValueError:
            pass

        # Return as string
        return value

    def _parse_legacy_format(self, content: str) -> List[XMLToolCall]:
        """
        Parse legacy XML tool formats for backwards compatibility.
        This handles formats like <tool_name>...</tool_name> or
        <tool_name param="value">...</tool_name>
        """
        tool_calls = []

        # Pattern for finding XML-like tags
        tag_pattern = re.compile(
            r'<([a-zA-Z][\w\-]*)((?:\s+[\w\-]+=["\'][^"\']*["\'])*)\s*>(.*?)</\1>',
            re.DOTALL,
        )

        for match in tag_pattern.finditer(content):
            tag_name = match.group(1)
            attributes_str = match.group(2)
            inner_content = match.group(3)

            # Skip our own format tags
            if tag_name in ("function_calls", "invoke", "parameter"):
                continue

            parameters = {}
            parsing_details = {
                "format": "legacy",
                "tag_name": tag_name,
                "attributes": {},
                "inner_content": inner_content.strip(),
            }

            # Parse attributes
            if attributes_str:
                attr_pattern = re.compile(r'([\w\-]+)=["\']([^"\']*)["\']')
                for attr_match in attr_pattern.finditer(attributes_str):
                    attr_name = attr_match.group(1)
                    attr_value = attr_match.group(2)
                    parameters[attr_name] = self._parse_parameter_value(attr_value)
                    parsing_details["attributes"][attr_name] = attr_value

            # If there's inner content and no attributes, use it as a 'content' parameter
            if inner_content.strip() and not parameters:
                parameters["content"] = inner_content.strip()

            # Convert tag name to function name (e.g., create-file -> create_file)
            function_name = tag_name.replace("-", "_")

            tool_calls.append(
                XMLToolCall(
                    function_name=function_name,
                    parameters=parameters,
                    raw_xml=match.group(0),
                    parsing_details=parsing_details,
                )
            )

        return tool_calls

    def format_tool_call(self, function_name: str, parameters: Dict[str, Any]) -> str:
        """
        Format a tool call in the Cursor-style XML format.

        Args:
            function_name: Name of the function to call
            parameters: Dictionary of parameters

        Returns:
            Formatted XML string
        """
        lines = ["<function_calls>", '<invoke name="{}">'.format(function_name)]

        for param_name, param_value in parameters.items():
            # Convert value to string representation
            if isinstance(param_value, (dict, list)):
                value_str = json.dumps(param_value)
            elif isinstance(param_value, bool):
                value_str = str(param_value).lower()
            else:
                value_str = str(param_value)

            lines.append(
                '<parameter name="{}">{}</parameter>'.format(param_name, value_str)
            )

        lines.extend(["</invoke>", "</function_calls>"])
        return "\n".join(lines)

    def validate_tool_call(
        self, tool_call: XMLToolCall, expected_params: Optional[Dict[str, type]] = None
    ) -> Tuple[bool, Optional[str]]:
        """
        Validate a tool call against expected parameters.

        Args:
            tool_call: The XMLToolCall to validate
            expected_params: Optional dict of parameter names to expected types

        Returns:
            Tuple of (is_valid, error_message)
        """
        if not tool_call.function_name:
            return False, "Function name is required"

        if expected_params:
            for param_name, expected_type in expected_params.items():
                if param_name not in tool_call.parameters:
                    return False, f"Missing required parameter: {param_name}"

                param_value = tool_call.parameters[param_name]
                if not isinstance(param_value, expected_type):
                    return (
                        False,
                        f"Parameter {param_name} should be of type {expected_type.__name__}",
                    )

        return True, None


# Convenience function for quick parsing
def parse_xml_tool_calls(content: str, strict_mode: bool = False) -> List[XMLToolCall]:
    """
    Parse XML tool calls from content.

    Args:
        content: The text content potentially containing XML tool calls
        strict_mode: If True, only accept the Cursor-style format

    Returns:
        List of parsed XMLToolCall objects
    """
    parser = XMLToolParser(strict_mode=strict_mode)
    return parser.parse_content(content)


if __name__ == "__main__":
    # Example usage
    content = """
    <function_calls>
    <invoke name="create_file">
    <parameter name="path">["/path/to/file.txt","/path/to/file2.txt"]</parameter>
    <parameter name="content">Hello, world!</parameter>
    </invoke>
    <invoke name="read_file">
    <parameter name="path">/path/to/file.txt</parameter>
    </invoke>
    </function_calls>
    
    <function_calls>
    <analysis>最终判断依据及结论</analysis>
    <invoke name="write_file">
    <parameter name="path">["/path/to/file.txt"]</parameter>
    <parameter name="content">Hello, world!</parameter>
    </invoke>
    </function_calls>
    
    """
    # Parse the content
    tool_calls = parse_xml_tool_calls(content)  # strict_mode=True)
    # Print the parsed tool calls
    for call in tool_calls:
        print(f"Function: {call.function_name}")
        print(f"Parameters: {call.parameters}")
        print(f"Raw XML: {call.raw_xml}")
        print(f"Parsing Details: {call.parsing_details}")
        print("-" * 40)



import logging
from os import getenv
from typing import Literal, Optional, Sequence

from duowen_agent.error import ToolError
from duowen_agent.tools.base import BaseTool
from pydantic import BaseModel, Field
from tavily import TavilyClient

from .entity import ToolSearchResult, ToolSearchResultDetails


class TavilyParameters(BaseModel):
    """Parameters definition for Tavily Search API"""

    query: str = Field(description="用户需要搜索的查询字符串")
    search_depth: Literal["basic", "advanced"] = Field(
        default="basic",
        description="搜索深度模式，basic=基础搜索（快速响应），advanced=高级搜索（更深入分析）",
    )
    topic: Literal["general", "news"] = Field(
        default="general", description="搜索主题类型，general=综合信息，news=时效性新闻"
    )
    days: int = Field(
        default=3,
        ge=1,
        description="搜索结果时间范围（仅限news主题），单位天数，默认最近3天",
    )
    max_results: int = Field(
        default=5, ge=1, le=20, description="最大返回结果数量（1-20之间）"
    )
    include_domains: Optional[Sequence[str]] = Field(
        default=None, description="指定需要搜索的域名列表（如['wikipedia.org']）"
    )
    exclude_domains: Optional[Sequence[str]] = Field(
        default=None, description="需要排除的域名列表（如['twitter.com']）"
    )
    include_answer: bool = Field(
        default=False, description="是否在结果中包含AI生成的答案摘要"
    )
    include_raw_content: bool = Field(
        default=False, description="是否包含网页原始内容（可能影响响应速度）"
    )
    include_images: bool = Field(default=False, description="是否在结果中返回图片链接")


class Tavily(BaseTool):

    name: str = "tavily-search-api"
    description: str = (
        "专业网络搜索引擎，支持深度内容抓取和定制过滤。"
        "优势：精准的来源控制、时效性筛选、结构化数据输出。"
        "适用场景：新闻追踪、学术研究、市场分析等需要最新网络信息的任务。"
    )
    parameters = TavilyParameters

    def __init__(
        self, api_key: str = None, max_results: int = 5, token_limit: int = 8000
    ):
        super().__init__()
        self.tavily_client = TavilyClient(api_key or getenv("TAVILY_API_KEY"))
        self.token_limit = token_limit
        self.max_results = max_results

    def search(
        self,
        query: str,
        search_depth: Literal["basic", "advanced"] = "advanced",
        **kwargs,
    ) -> ToolSearchResult:
        try:
            response = self.tavily_client.search(
                query=query, search_depth=search_depth, **kwargs
            )
            data1 = []
            _token_limit = self.token_limit
            for res in response["results"][: self.max_results]:
                _res = ToolSearchResultDetails(
                    url=res["url"],
                    title=res["title"],
                    content=res["content"],
                )
                if _res.chat_token <= _token_limit:
                    data1.append(_res)
                    _token_limit -= _res.chat_token
                else:
                    break

            return ToolSearchResult(result=data1)
        except Exception as e:
            logging.exception(e)
            raise ToolError(str(e))

    def _run(
        self,
        query: str,
        search_depth: Literal["basic", "advanced"] = "advanced",
        **kwargs,
    ):
        _data = self.search(query, **kwargs)
        return _data.content_with_weight, _data



import json
from typing import Literal, List, Union, Dict

from pydantic import BaseModel, Field

from duowen_agent.utils.core_utils import remove_think

openai_params_list = {
    "messages",
    "model",
    "audio",
    "frequency_penalty",
    "function_call",
    "functions",
    "logit_bias",
    "logprobs",
    "max_completion_tokens",
    "max_tokens",
    "metadata",
    "modalities",
    "n",
    "parallel_tool_calls",
    "prediction",
    "presence_penalty",
    "response_format",
    "seed",
    "service_tier",
    "stop",
    "store",
    "stream",
    "stream_options",
    "temperature",
    "tool_choice",
    "tools",
    "top_logprobs",
    "top_p",
    "user",
}


class BaseContent(BaseModel):
    """内容基类"""


class TextContent(BaseContent):
    type: Literal["text"] = "text"
    text: str

    def to_dict(self):
        return {"type": self.type, "text": self.text}


class ImageURLContent(BaseContent):
    type: Literal["image_url"] = "image_url"
    image_url: Dict[str, str]

    def to_dict(self):
        return {"type": self.type, "image_url": self.image_url}


ContentUnion = Union[TextContent, ImageURLContent]


class Message(BaseModel):
    role: Literal["system", "user", "assistant"] = "user"
    content: Union[str, List[ContentUnion]]

    def __init__(
        self,
        content: Union[str, List[ContentUnion]],
        role: Literal["system", "user", "assistant"] = "user",
    ):
        super().__init__(content=content, role=role)

    def __getitem__(self, item):
        if item == "content":
            return self.content
        elif item == "role":
            return self.role
        else:
            raise KeyError(f"Message has no key {item}")

    def format_str(self) -> str:
        if isinstance(self.content, str):
            return (
                f"<{self.role}>\n"
                + "\n".join(["  " + j for j in self.content.split("\n")])
                + f"\n</{self.role}>"
            )
        else:
            return (
                f"<{self.role}>\n"
                + "\n".join([f"  {str(j)}" for j in self.content])
                + f"\n</{self.role}>"
            )

    def to_dict(self) -> dict:

        if isinstance(self.content, str):
            return {"role": self.role, "content": self.content}
        else:
            return {"role": self.role, "content": [i.to_dict() for i in self.content]}


class SystemMessage(Message):
    def __init__(self, content: Union[str, List[ContentUnion]]):
        super().__init__(content, "system")


class UserMessage(Message):
    def __init__(self, content: Union[str, List[ContentUnion]]):
        super().__init__(content, "user")


class AssistantMessage(Message):
    def __init__(self, content: Union[str, List[ContentUnion]]):
        super().__init__(content, "assistant")


class MessagesSet(BaseModel):
    message_list: List[Message] = []

    def __init__(self, message_list: List[dict] | List[Message] = None):
        if message_list:
            if isinstance(message_list[0], dict):
                message_list = [Message(**i) for i in message_list]
            elif isinstance(message_list[0], Message):
                pass
            else:
                raise ValueError("MessagesSet init message_list type error")
            super().__init__(message_list=message_list)
        else:
            super().__init__()

    def remove_assistant_think(self):
        """推理模型需要剔除think部分"""
        for message in self.message_list:
            if message.role == "assistant":
                message.content = remove_think(message.content)
        return self

    def init_message_list(self, message: List[Dict[str, str]]):
        for i in message:
            if i["role"] == "assistant":
                self.add_assistant(i["content"])
            elif i["role"] == "system":
                self.add_system(i["content"])
            elif i["role"] == "user":
                self.add_user(i["content"])
        return self

    def add_user(self, content: Union[str, ContentUnion, List[ContentUnion]]):
        self.message_list.append(UserMessage(content))
        return self

    def add_assistant(self, content: Union[str, ContentUnion, List[ContentUnion]]):
        self.message_list.append(AssistantMessage(content))
        return self

    def add_system(self, content: Union[str, ContentUnion, List[ContentUnion]]):
        self.message_list.append(SystemMessage(content))
        return self

    def append_messages(
        self, messages_set: Union["MessagesSet", List[UserMessage | AssistantMessage]]
    ):
        if type(messages_set) is MessagesSet:
            self.message_list = self.message_list + messages_set.message_list
        else:
            for message in messages_set:
                if type(message) is Message:
                    self.message_list.append(message)
                else:
                    raise ValueError("MessagesSet append_messages type error")
        return self

    def get_messages(self):
        return [i.to_dict() for i in self.message_list]

    def get_format_messages(self):
        _data = []

        for i in self.message_list:
            _data.append(i.format_str())

        return "\n\n".join(_data)

    def pretty_print(self):
        print(self.get_format_messages())

    def get_last_message(self):
        """Returns the last message if available, otherwise returns None."""
        if not self.message_list:
            return None
        return self.message_list[-1]

    def __add__(self, other: "MessagesSet") -> "MessagesSet":
        if not isinstance(other, MessagesSet):
            raise TypeError("Can only add MessagesSet to MessagesSet")
        return MessagesSet(self.message_list + other.message_list)

    def __iadd__(self, other: "MessagesSet") -> "MessagesSet":
        if not isinstance(other, MessagesSet):
            raise TypeError("Can only add MessagesSet to MessagesSet")
        return MessagesSet(self.message_list + other.message_list)

    def __getitem__(self, item):
        return self.message_list[item]

    def __len__(self):
        return len(self.message_list)

    def __bool__(self):
        return bool(self.message_list)

    def __iter__(self):
        for item in self.message_list:
            yield item

    def __repr__(self):
        return f"MessagesSet({self.message_list})"

    def __str__(self):
        return f"MessagesSet({str(self.message_list)[:200]})"


class Tool(BaseModel):
    name: str
    arguments: Dict = Field(default_factory=dict)
    think: str = None

    def __str__(self):
        return json.dumps(
            {"name": self.name, "arguments": self.arguments, "think": self.think},
            ensure_ascii=False,
        )


class ToolsCall(BaseModel):
    think: str = None
    tools: List[Tool] = Field(default_factory=list)

    def __str__(self):
        return json.dumps(
            {"think": self.think, "tools": [i.model_dump() for i in self.tools]},
            ensure_ascii=False,
        )



from duowen_agent.utils.core_utils import stream_to_string

META_PROMPT = """
Given a current prompt and a change description, produce a detailed system prompt to guide a language model in completing the task effectively.

Your final output will be the full corrected prompt verbatim. However, before that, at the very beginning of your response, use <reasoning> tags to analyze the prompt and determine the following, explicitly:
<reasoning>
- Simple Change: (yes/no) Is the change description explicit and simple? (If so, skip the rest of these questions.)
- Reasoning: (yes/no) Does the current prompt use reasoning, analysis, or chain of thought? 
    - Identify: (max 10 words) if so, which section(s) utilize reasoning?
    - Conclusion: (yes/no) is the chain of thought used to determine a conclusion?
    - Ordering: (before/after) is the chain of though located before or after 
- Structure: (yes/no) does the input prompt have a well defined structure
- Examples: (yes/no) does the input prompt have few-shot examples
    - Representative: (1-5) if present, how representative are the examples?
- Complexity: (1-5) how complex is the input prompt?
    - Task: (1-5) how complex is the implied task?
    - Necessity: ()
- Specificity: (1-5) how detailed and specific is the prompt? (not to be confused with length)
- Prioritization: (list) what 1-3 categories are the MOST important to address.
- Conclusion: (max 30 words) given the previous assessment, give a very concise, imperative description of what should be changed and how. this does not have to adhere strictly to only the categories listed
</reasoning>

# Guidelines

- Understand the Task: Grasp the main objective, goals, requirements, constraints, and expected output.
- Minimal Changes: If an existing prompt is provided, improve it only if it's simple. For complex prompts, enhance clarity and add missing elements without altering the original structure.
- Reasoning Before Conclusions**: Encourage reasoning steps before any conclusions are reached. ATTENTION! If the user provides examples where the reasoning happens afterward, REVERSE the order! NEVER START EXAMPLES WITH CONCLUSIONS!
    - Reasoning Order: Call out reasoning portions of the prompt and conclusion parts (specific fields by name). For each, determine the ORDER in which this is done, and whether it needs to be reversed.
    - Conclusion, classifications, or results should ALWAYS appear last.
- Examples: Include high-quality examples if helpful, using placeholders [in brackets] for complex elements.
   - What kinds of examples may need to be included, how many, and whether they are complex enough to benefit from placeholders.
- Clarity and Conciseness: Use clear, specific language. Avoid unnecessary instructions or bland statements.
- Formatting: Use markdown features for readability. DO NOT USE ``` CODE BLOCKS UNLESS SPECIFICALLY REQUESTED.
- Preserve User Content: If the input task or prompt includes extensive guidelines or examples, preserve them entirely, or as closely as possible. If they are vague, consider breaking down into sub-steps. Keep any details, guidelines, examples, variables, or placeholders provided by the user.
- Constants: DO include constants in the prompt, as they are not susceptible to prompt injection. Such as guides, rubrics, and examples.
- Output Format: Explicitly the most appropriate output format, in detail. This should include length and syntax (e.g. short sentence, paragraph, JSON, etc.)
    - For tasks outputting well-defined or structured data (classification, JSON, etc.) bias toward outputting a JSON.
    - JSON should never be wrapped in code blocks (```) unless explicitly requested.

The final prompt you output should adhere to the following structure below. Do not include any additional commentary, only output the completed system prompt. SPECIFICALLY, do not include any additional messages at the start or end of the prompt. (e.g. no "---")

[Concise instruction describing the task - this should be the first line in the prompt, no section header]

[Additional details as needed.]

[Optional sections with headings or bullet points for detailed steps.]

# Steps [optional]

[optional: a detailed breakdown of the steps necessary to accomplish the task]

# Output Format

[Specifically call out how the output should be formatted, be it response length, structure e.g. JSON, markdown, etc]

# Examples [optional]

[Optional: 1-3 well-defined examples with placeholders if necessary. Clearly mark where examples start and end, and what the input and output are. User placeholders as necessary.]
[If the examples are shorter than what a realistic example is expected to be, make a reference with () explaining how real examples should be longer / shorter / different. AND USE PLACEHOLDERS! ]

# Notes [optional]

[optional: edge cases, details, and an area to call or repeat out specific important considerations]
[NOTE: you must start with a <reasoning> section. the immediate next token you produce should be <reasoning>]
""".strip()

from duowen_agent.llm.chat_model import OpenAIChat


def generate_prompt(llm: OpenAIChat, task_or_prompt: str) -> str:
    return stream_to_string(
        llm.chat_for_stream(
            messages=[
                {
                    "role": "system",
                    "content": META_PROMPT,
                },
                {
                    "role": "user",
                    "content": "Task, Goal, or Current Prompt:\n" + task_or_prompt,
                },
            ]
        )
    )


SCHEMA_META_PROMPT = """
# Instructions
Return a valid schema for the described JSON.

You must also make sure:
- all fields in an object are set as required
- I REPEAT, ALL FIELDS MUST BE MARKED AS REQUIRED
- all objects must have additionalProperties set to false
    - because of this, some cases like "attributes" or "metadata" properties that would normally allow additional properties should instead have a fixed set of properties
- all objects must have properties defined
- field order matters. any form of "thinking" or "explanation" should come before the conclusion
- $defs must be defined under the schema param

Notable keywords NOT supported include:
- For strings: minLength, maxLength, pattern, format
- For numbers: minimum, maximum, multipleOf
- For objects: patternProperties, unevaluatedProperties, propertyNames, minProperties, maxProperties
- For arrays: unevaluatedItems, contains, minContains, maxContains, minItems, maxItems, uniqueItems

Other notes:
- definitions and recursion are supported
- only if necessary to include references e.g. "$defs", it must be inside the "schema" object

# Examples
Input: Generate a math reasoning schema with steps and a final answer.
Output: {
    "name": "math_reasoning",
    "type": "object",
    "properties": {
        "steps": {
            "type": "array",
            "description": "A sequence of steps involved in solving the math problem.",
            "items": {
                "type": "object",
                "properties": {
                    "explanation": {
                        "type": "string",
                        "description": "Description of the reasoning or method used in this step."
                    },
                    "output": {
                        "type": "string",
                        "description": "Result or outcome of this specific step."
                    }
                },
                "required": [
                    "explanation",
                    "output"
                ],
                "additionalProperties": false
            }
        },
        "final_answer": {
            "type": "string",
            "description": "The final solution or answer to the math problem."
        }
    },
    "required": [
        "steps",
        "final_answer"
    ],
    "additionalProperties": false
}

Input: Give me a linked list
Output: {
    "name": "linked_list",
    "type": "object",
    "properties": {
        "linked_list": {
            "$ref": "#/$defs/linked_list_node",
            "description": "The head node of the linked list."
        }
    },
    "$defs": {
        "linked_list_node": {
            "type": "object",
            "description": "Defines a node in a singly linked list.",
            "properties": {
                "value": {
                    "type": "number",
                    "description": "The value stored in this node."
                },
                "next": {
                    "anyOf": [
                        {
                            "$ref": "#/$defs/linked_list_node"
                        },
                        {
                            "type": "null"
                        }
                    ],
                    "description": "Reference to the next node; null if it is the last node."
                }
            },
            "required": [
                "value",
                "next"
            ],
            "additionalProperties": false
        }
    },
    "required": [
        "linked_list"
    ],
    "additionalProperties": false
}

Input: Dynamically generated UI
Output: {
    "name": "ui",
    "type": "object",
    "properties": {
        "type": {
            "type": "string",
            "description": "The type of the UI component",
            "enum": [
                "div",
                "button",
                "header",
                "section",
                "field",
                "form"
            ]
        },
        "label": {
            "type": "string",
            "description": "The label of the UI component, used for buttons or form fields"
        },
        "children": {
            "type": "array",
            "description": "Nested UI components",
            "items": {
                "$ref": "#"
            }
        },
        "attributes": {
            "type": "array",
            "description": "Arbitrary attributes for the UI component, suitable for any element",
            "items": {
                "type": "object",
                "properties": {
                    "name": {
                        "type": "string",
                        "description": "The name of the attribute, for example onClick or className"
                    },
                    "value": {
                        "type": "string",
                        "description": "The value of the attribute"
                    }
                },
                "required": [
                    "name",
                    "value"
                ],
                "additionalProperties": false
            }
        }
    },
    "required": [
        "type",
        "label",
        "children",
        "attributes"
    ],
    "additionalProperties": false
}
""".strip()


def generate_schema(llm: OpenAIChat, description: str) -> str:
    return stream_to_string(
        llm.chat_for_stream(
            messages=[
                {
                    "role": "system",
                    "content": SCHEMA_META_PROMPT,
                },
                {
                    "role": "user",
                    "content": "Description:\n" + description,
                },
            ],
        )
    )



from typing import List

from pydantic import BaseModel, Field

from duowen_agent.llm.chat_model import BaseAIChat
from duowen_agent.llm.tokenizer import tokenizer
from .base import BaseComponent
from ...prompt.prompt_build import GeneralPromptBuilder
from ...utils.core_utils import json_observation, stream_to_string


class Ranker(BaseComponent):
    """通过语言模型实现 rerank能力 不支持分值，只能排序"""

    # query: str, documents
    def __init__(self, llm: BaseAIChat):
        super().__init__()
        self.llm = llm
        self.content_tokens_limit = None
        self.documents = None
        self.question_tokens = None
        self.prompt_tokens = 1000
        self.rank_limit = 5
        self.query = None

    def init_data(self, query: str, documents: List[str], rank_limit=5):
        self.query = query
        self.rank_limit = rank_limit
        self.question_tokens = tokenizer.chat_len(query)
        self.documents = documents
        self.content_tokens_limit = (
            self.llm.token_limit - self.prompt_tokens - self.question_tokens
        )

    def cut_passages(self):
        _content_tokens = self.content_tokens_limit
        _passages = []
        for _chunk in self.documents:
            _curr_token = tokenizer.chat_len(_chunk)
            _content_tokens = _content_tokens - _curr_token
            if _content_tokens > 0:
                _passages.append(_chunk)
            else:
                break
        self.documents = _passages

    def chk_passages_tokens_limit(self):
        if tokenizer.chat_len("".join(self.documents)) > self.content_tokens_limit:
            return False
        else:
            return True

    def get_prefix_prompt(self, num):
        return [
            {
                "role": "system",
                "content": "You are RankGPT, an intelligent assistant that can rank passages based on their relevancy to the query.",
            },
            {
                "role": "user",
                "content": f"I will provide you with {num} passages, each indicated by number identifier []. \nRank the passages based on their relevance to query: {self.query}.",
            },
            {"role": "assistant", "content": "Okay, please provide the passages."},
        ]

    def get_post_prompt(self, num):
        return f"Search Query: {self.query}. \nRank the {num} passages above based on their relevance to the search query. The passages should be listed in descending order using identifiers. The most relevant passages should be listed first. The output format should be [] > [], e.g., [1] > [2]. Only response the ranking results, do not say any word or explain."

    def create_permutation_instruction(self):
        if not self.chk_passages_tokens_limit():
            raise ValueError(
                f"Agent Ranker token passages overly long, model tokens limit number {self.llm.token_limit}."
            )
        num = len(self.documents)
        messages = self.get_prefix_prompt(num)
        rank = 0
        for hit in self.documents:
            rank += 1
            content = hit
            content = content.replace("Title: Content: ", "")
            content = content.strip()
            messages.append({"role": "user", "content": f"[{rank}] {content}"})
            messages.append(
                {"role": "assistant", "content": f"Received passage [{rank}]."}
            )
        messages.append({"role": "user", "content": self.get_post_prompt(num)})
        return messages

    def run_llm(self, messages):
        response = stream_to_string(
            self.llm.chat_for_stream(messages=messages, temperature=0)
        )
        return response

    async def arun_llm(self, messages):
        response = await self.llm.achat(messages=messages, temperature=0)
        return response

    @staticmethod
    def clean_response(response: str):
        new_response = ""
        for c in response:
            if not c.isdigit():
                new_response += " "
            else:
                new_response += c
        new_response = new_response.strip()
        return new_response

    @staticmethod
    def remove_duplicate(response):
        new_response = []
        for c in response:
            if c not in new_response:
                new_response.append(c)
        return new_response

    def receive_permutation(self, permutation):
        _passages = []
        response = self.clean_response(permutation)
        response = [int(x) - 1 for x in response.split()]
        response = self.remove_duplicate(response)
        original_rank = [tt for tt in range(len(self.documents))]
        response = [ss for ss in response if ss in original_rank]
        response = response + [tt for tt in original_rank if tt not in response]
        for x in response[: self.rank_limit]:
            _passages.append(self.documents[x])
        return _passages

    def run(self, query: str, documents: List[str], rank_limit=5) -> List[str]:
        self.init_data(query, documents, rank_limit)
        if self.rank_limit < len(self.documents):
            messages = self.create_permutation_instruction()
            permutation = self.run_llm(messages)
            item = self.receive_permutation(permutation)
            return item
        else:
            return self.documents

    async def arun(self, query: str, documents: List[str], rank_limit=5) -> List[str]:
        self.init_data(query, documents, rank_limit)
        if self.rank_limit < len(self.documents):
            messages = self.create_permutation_instruction()
            permutation = await self.arun_llm(messages)
            item = self.receive_permutation(permutation)
            return item
        else:
            return self.documents


class RelevanceScore(BaseModel):
    reasoning: str = Field(..., description="推理过程(50字以内)")
    relevance_score: int = Field(..., ge=0, le=10, description="相关性分数")


class RankerScore(BaseComponent):
    """通过语言模型实现 对文档进行相关性打分"""

    def __init__(self, llm: BaseAIChat):
        super().__init__()
        self.llm = llm

    def build_prompt(self) -> GeneralPromptBuilder:
        return GeneralPromptBuilder(
            instruction="""根据文档片段与用户问题的相关性进行1-10分的评分，并严格遵守评分标准。确保所有结论基于详细的分析推理得出，推理过程必须出现在最终评分之前。

### 评分标准
对每个评分级别定义具体要求，需严格遵循：
1. **完全不相关** (1分)：文档与问题主题毫无关联
2. **极弱相关** (2分)：仅有个别非关键单词相同
3. **微弱相关** (3分)：涉及相同领域但无实质信息
4. **部分相关** (4分)：解决次要问题或提供间接信息
5. **基本相关** (5分)：能回应部分核心问题但存在重大缺失
6. **中等相关** (6分)：覆盖主要问题但需用户自行补充
7. **良好相关** (7分)：直接回答核心问题但有轻微遗漏
8. **高度相关** (8分)：精确解答主要问题并有额外支持
9. **极强相关** (9分)：完全解决问题并给出深度洞察
10. **完美相关** (10分)：精准确认所有细节并预判延伸需求
""",
            step="""1. **识别关键要素**
   - 提取问题核心需求：[主要查询点]
   - 标记文档核心信息：[关键知识点]
2. **对比分析**
   - 主题匹配度：[领域/场景重合度]
   - 信息完整度：[问题需求覆盖比例]
   - 细节精确度：[数据/术语准确率]
3. **缺失项检测**
   - 未覆盖的需求点：[列出具体遗漏]
   - 矛盾信息：[存在冲突的陈述]
4. **推理结论**
   - 说明评分依据与标准条款对应关系""",
            output_format=RelevanceScore,
            sample="""**输入组合：**
用户问题："网站显示500错误怎么办？"
文档片段："HTTP状态码200表示请求成功，404表示资源未找到"

**输出：**
```json
{
  "reasoning": "由于提供的回答内容完全未触及用户核心问题——服务器500错误的原因或解决方案，且相关文档仅涵盖200/404状态码而明确缺失对5xx错误的说明，依据关键主题匹配度（0%）与核心需求覆盖度（0%）的分析结果，根据评分标准第1条：属于完全无关主题。",
  "score": 1
}
```

**(实际案例中，文档与问题应存在部分关联时，推理段落需更详细展开具体匹配点分析)**""",
            note="""- 当问题包含多子需求时，逐点分析关联性
- 对专业术语需验证使用一致性
- 文档隐含信息（如推论结论）不计入评分依据
- 分值必须精确对应评分标准条目""",
        )

    def run(self, query: str, document: str) -> int:

        _prompt = self.build_prompt().get_instruction(
            f"用户问题：{query}\n\n文档片段：\n{document}"
        )

        resp = stream_to_string(self.llm.chat_for_stream(_prompt))

        resp = json_observation(resp, RelevanceScore)

        return resp.relevance_score

    async def arun(self, query: str, document: str) -> int:
        _prompt = self.build_prompt().get_instruction(
            f"用户问题：{query}\n\n文档片段：\n{document}"
        )
        resp = await self.llm.achat(_prompt)
        resp: RelevanceScore = json_observation(resp, RelevanceScore)
        return resp.relevance_score



import html
import json
import logging
import math  # 导入 math 库
import numbers
import os
import re
from datetime import datetime
from hashlib import md5
from typing import Any, Union, List

import networkx as nx
from networkx.readwrite import json_graph
from pyvis.network import Network

from duowen_agent.llm import tokenizer


def create_file_dir(file_name):
    dir_name = os.path.dirname(file_name)
    if dir_name and not os.path.exists(dir_name):
        os.makedirs(dir_name, exist_ok=True)


def compute_args_hash(*args):
    return md5(str(args).encode()).hexdigest()


def print_call_back(msg: str):
    print(f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}: {msg}")


def is_interrupt():
    """集成系统需要实现的中断函数"""
    return False


def clean_str(input: Any) -> str:
    """Clean an input string by removing HTML escapes, control characters, and other unwanted characters."""
    # If we get non-string input, just give it back
    if not isinstance(input, str):
        return input

    result = html.unescape(input.strip())
    # https://stackoverflow.com/questions/4324790/removing-control-characters-from-a-string-in-python
    return re.sub(r"[\"\x00-\x1f\x7f-\x9f]", "", result)


def is_float_regex(value):
    return bool(re.match(r"^[-+]?[0-9]*\.?[0-9]+$", value))


def flat_uniq_list(arr, key):
    res = []
    for a in arr:
        a = a[key]
        if isinstance(a, list):
            res.extend(a)
        else:
            res.append(a)
    return list(set(res))


def truncate(string: str, max_len: int) -> str:
    return tokenizer.truncate_emb(string, max_len)


def tidy_graph(graph: nx.Graph, callback, check_attribute: bool = True):
    """
    Ensure all nodes and edges in the graph have some essential attribute.
    """

    def is_valid_item(node_attrs: dict) -> bool:
        valid_node = True
        for attr in ["description", "source_id"]:
            if attr not in node_attrs:
                valid_node = False
                break
        return valid_node

    if check_attribute:
        purged_nodes = []
        for node, node_attrs in graph.nodes(data=True):
            if not is_valid_item(node_attrs):
                purged_nodes.append(node)
        for node in purged_nodes:
            graph.remove_node(node)
        if purged_nodes and callback:
            callback(f"由于缺少基本属性，从图中清除 {len(purged_nodes)} 节点")

    purged_edges = []
    for source, target, attr in graph.edges(data=True):
        if check_attribute:
            if not is_valid_item(attr):
                purged_edges.append((source, target))

    for source, target in purged_edges:
        graph.remove_edge(source, target)
    if purged_edges and callback:
        callback(f"由于缺少基本属性，从图中清除 {len(purged_edges)} 边。")


def dump_graph(graph: nx.Graph):
    return json.dumps(
        nx.node_link_data(graph, edges="edges"), ensure_ascii=False, indent=2
    )


def load_json_graph(content):
    return json_graph.node_link_graph(json.loads(content), edges="edges")


def get_from_to(node1, node2):
    if node1 < node2:
        return node1, node2
    else:
        return node2, node1


def dict_has_keys_with_types(
    data: dict, expected_fields: list[tuple[str, type]]
) -> bool:
    """Return True if the given dictionary has the given keys with the given types."""
    for field, field_type in expected_fields:
        if field not in data:
            return False

        value = data[field]
        if not isinstance(value, field_type):
            return False
    return True


def parse_value(value: str):
    """Convert a string value to its appropriate type (int, float, bool, None, or keep as string). Work as a more broad 'eval()'"""
    value = value.strip()

    if value == "null":
        return None
    elif value == "true":
        return True
    elif value == "false":
        return False
    else:
        # Try to convert to int or float
        try:
            if "." in value:  # If there's a dot, it might be a float
                return float(value)
            else:
                return int(value)
        except ValueError:
            # If conversion fails, return the value as-is (likely a string)
            return value.strip('"')  # Remove surrounding quotes if they exist


def extract_first_complete_json(s: str):
    """Extract the first complete JSON object from the string using a stack to track braces."""
    stack = []
    first_json_start = None

    for i, char in enumerate(s):
        if char == "{":
            stack.append(i)
            if first_json_start is None:
                first_json_start = i
        elif char == "}":
            if stack:
                start = stack.pop()
                if not stack:
                    first_json_str = s[first_json_start : i + 1]
                    try:
                        # Attempt to parse the JSON string
                        return json.loads(first_json_str.replace("\n", ""))
                    except json.JSONDecodeError as e:
                        logging.error(
                            f"JSON decoding failed: {e}. Attempted string: {first_json_str[:50]}..."
                        )
                        return None
                    finally:
                        first_json_start = None
    logging.warning("No complete JSON object found in the input string.")
    return None


def extract_values_from_json(json_string):
    """Extract key values from a non-standard or malformed JSON string, handling nested objects."""
    extracted_values = {}

    # Enhanced pattern to match both quoted and unquoted values, as well as nested objects
    regex_pattern = r'(?P<key>"?\w+"?)\s*:\s*(?P<value>{[^}]*}|".*?"|[^,}]+)'

    for match in re.finditer(regex_pattern, json_string, re.DOTALL):
        key = match.group("key").strip('"')  # Strip quotes from key
        value = match.group("value").strip()

        # If the value is another nested JSON (starts with '{' and ends with '}'), recursively parse it
        if value.startswith("{") and value.endswith("}"):
            extracted_values[key] = extract_values_from_json(value)
        else:
            # Parse the value into the appropriate type (int, float, bool, etc.)
            extracted_values[key] = parse_value(value)

    if not extracted_values:
        logging.warning("No values could be extracted from the string.")

    return extracted_values


def convert_response_to_json(response: str) -> dict:
    """Convert response string to JSON, with error handling and fallback to non-standard JSON extraction."""
    prediction_json = extract_first_complete_json(response)

    if prediction_json is None:
        logging.info("Attempting to extract values from a non-standard JSON string...")
        prediction_json = extract_values_from_json(response)

    if not prediction_json:
        logging.error("Unable to extract meaningful data from the response.")
    else:
        logging.info("JSON data successfully extracted.")

    return prediction_json


def truncate_list_by_token_size(
    list_data: list,
    key: callable,
    max_token_size: int,
):
    """Truncate a list of data by token size using a provided tokenizer wrapper."""
    if max_token_size <= 0:
        return []
    tokens = 0
    for i, data in enumerate(list_data):
        tokens += (
            len(tokenizer.chat_encode(key(data))) + 1
        )  # 防御性，模拟通过\n拼接列表的情况
        if tokens > max_token_size:
            return list_data[:i]
    return list_data


def enclose_string_with_quotes(content: Any) -> str:
    """Enclose a string with quotes"""
    if isinstance(content, numbers.Number):
        return str(content)
    content = str(content)
    content = content.strip().strip("'").strip('"')
    return f'"{content}"'


def list_of_list_to_csv(data: list[list]):
    return "\n".join(
        [
            ",\t".join([f"{enclose_string_with_quotes(data_dd)}" for data_dd in data_d])
            for data_d in data
        ]
    )


def split_string_by_multi_markers(content: str, markers: list[str]) -> list[str]:
    """Split a string by multiple markers"""
    if not markers:
        return [content]
    results = re.split("|".join(re.escape(marker) for marker in markers), content)
    return [r.strip() for r in results if r.strip()]


class NetworkXUtils:

    def __init__(self, graph: nx.Graph | str = None):
        if isinstance(graph, str):
            self._graph = self.loads(graph)
        elif isinstance(graph, nx.Graph):
            self._graph = graph
        elif graph is None:
            self._graph = nx.Graph()
        else:
            raise ValueError("graph must be a nx.Graph instance or a json string")

    def has_node(self, node_id: str) -> bool:
        return self._graph.has_node(node_id)

    def has_edge(self, source_node_id: str, target_node_id: str) -> bool:
        return self._graph.has_edge(source_node_id, target_node_id)

    def get_node(self, node_id: str) -> Union[dict, None]:
        return self._graph.nodes.get(node_id)

    def number_of_nodes(self) -> int:
        return self._graph.number_of_nodes()

    def number_of_edges(self) -> int:
        return self._graph.number_of_edges()

    def graph_density(self) -> float:
        return nx.density(self._graph)

    def get_nodes_batch(self, node_ids: list[str]) -> list[Union[dict, None]]:
        return [self.get_node(node_id) for node_id in node_ids]

    def node_degree(self, node_id: str) -> int:
        # [numberchiffre]: node_id not part of graph returns `DegreeView({})` instead of 0
        return self._graph.degree(node_id) if self._graph.has_node(node_id) else 0

    def node_degrees_batch(self, node_ids: List[str]) -> List[int]:
        return [self.node_degree(node_id) for node_id in node_ids]

    def edge_degree(self, src_id: str, tgt_id: str) -> int:
        return (self._graph.degree(src_id) if self._graph.has_node(src_id) else 0) + (
            self._graph.degree(tgt_id) if self._graph.has_node(tgt_id) else 0
        )

    def edge_degrees_batch(self, edge_pairs: list[tuple[str, str]]) -> list[int]:
        return [self.edge_degree(src_id, tgt_id) for src_id, tgt_id in edge_pairs]

    def get_edge(self, source_node_id: str, target_node_id: str) -> Union[dict, None]:
        return self._graph.edges.get((source_node_id, target_node_id))

    def get_edges_batch(
        self, edge_pairs: list[tuple[str, str]]
    ) -> list[Union[dict, None]]:
        return [
            self.get_edge(source_node_id, target_node_id)
            for source_node_id, target_node_id in edge_pairs
        ]

    def get_node_edges(self, source_node_id: str):
        if self._graph.has_node(source_node_id):
            return list(self._graph.edges(source_node_id))
        return None

    def get_nodes_edges_batch(self, node_ids: list[str]) -> list[list[tuple[str, str]]]:
        return [self.get_node_edges(node_id) for node_id in node_ids]

    def get_neighbors_graph(
        self,
        start_node: str,
        steps: int,
        top_k_neighbors: int = 100,  # 使用weight判断
        filter_entities: List[str] = None,
    ) -> nx.Graph:

        if not self.has_node(start_node):
            return nx.Graph()

        subgraph = nx.Graph()
        # Add the starting node to the subgraph
        subgraph.add_node(start_node, **self.get_node(start_node))

        q = [(start_node, 0)]
        visited_nodes = {start_node}

        while q:
            current_node, depth = q.pop(0)

            if depth >= steps:
                continue

            # Get neighbors and sort them for potential top-k pruning
            neighbors = list(self._graph.neighbors(current_node))

            if top_k_neighbors is not None:
                # Sort neighbors by edge attribute in descending order
                sorted_neighbors = sorted(
                    neighbors,
                    key=lambda n: self.get_edge(current_node, n).get("weight", 0),
                    reverse=True,
                )

                if filter_entities:
                    sorted_neighbors = [
                        n
                        for n in sorted_neighbors
                        if self.get_node(n).get("entity_type") in filter_entities
                    ]

                neighbors = sorted_neighbors[:top_k_neighbors]

            # sorted_neighbors = sorted(
            #     neighbors,
            #     key=lambda n: (
            #         self.get_node(n).get("pagerank", 0),
            #         self.get_node(n).get("rank", 0),
            #     ),
            #     reverse=True,
            # )[:top_k_node]

            for neighbor in neighbors:
                edge_data = self.get_edge(current_node, neighbor)
                node_data = self.get_node(neighbor)

                # Add node and edge to the subgraph
                if not subgraph.has_node(neighbor):
                    subgraph.add_node(neighbor, **node_data)

                if not subgraph.has_edge(current_node, neighbor):
                    subgraph.add_edge(current_node, neighbor, **edge_data)

                if neighbor not in visited_nodes:
                    visited_nodes.add(neighbor)
                    q.append((neighbor, depth + 1))

        return subgraph

    def upsert_node(self, node_id: str, node_data: dict[str, str]):
        self._graph.add_node(node_id, **node_data)

    def upsert_nodes_batch(self, nodes_data: list[tuple[str, dict[str, str]]]):
        for node_id, node_data in nodes_data:
            self.upsert_node(node_id, node_data)

    def upsert_edge(
        self, source_node_id: str, target_node_id: str, edge_data: dict[str, str]
    ):
        self._graph.add_edge(source_node_id, target_node_id, **edge_data)

    def upsert_edges_batch(self, edges_data: list[tuple[str, str, dict[str, str]]]):
        for source_node_id, target_node_id, edge_data in edges_data:
            self.upsert_edge(source_node_id, target_node_id, edge_data)

    def pagerank(self):
        pr = nx.pagerank(self._graph)
        for node_name, pagerank in pr.items():
            self._graph.nodes[node_name]["pagerank"] = pagerank
        return self._graph

    def dumps(self):
        return dump_graph(self._graph)

    @staticmethod
    def loads(content: str):
        return json_graph.node_link_graph(json.loads(content), edges="edges")


def similarity_node(
    nodes: list[str],
    embeddings: list[list[float]],
    sim_threshold=0.75,
    min_cluster_size=2,
    max_cluster_size=100,
) -> List[list[str]]:
    from sklearn.metrics.pairwise import cosine_similarity

    sim_matrix = cosine_similarity(embeddings)

    G = nx.Graph()
    for i in range(len(nodes)):
        G.add_node(i, text=nodes[i])
        for j in range(i + 1, len(nodes)):
            if sim_matrix[i][j] > sim_threshold:
                G.add_edge(i, j)

    clusters = list(nx.connected_components(G))
    return [
        [nodes[i] for i in list(i)]
        for i in clusters
        if max_cluster_size >= len(i) >= min_cluster_size
    ]


def create_styled_graph(g: nx.Graph, entity_color_map=None):
    """
    从一个不可修改的 networkx.Graph 对象创建一个带样式的、类似 Neo4j 的 pyvis 关系图。
    此版本适配了基于 entity_type 和 pagerank 的动态样式。

    Args:
        g (nx.Graph): 外部传入的 networkx 图对象。
        entity_color_map (dict): 实体类型到颜色的映射字典。
    """
    # 1. 创建一个 PyVis Network 对象，配置 Neo4j 风格
    is_directed = isinstance(g, nx.DiGraph)
    net = Network(
        height="800px",
        width="100%",
        bgcolor="#222222",
        font_color="white",
        notebook=False,
        select_menu=True,
        filter_menu=True,
        cdn_resources="in_line",
        directed=is_directed,
    )

    # 定义一个颜色映射表，用于根据 entity_type 分配颜色
    # 颜色经过精心挑选，以实现视觉上的和谐与区分度
    if entity_color_map is None:
        entity_color_map = {
            # --- 人物与社会组织 (暖色系：红、橙、黄) ---
            "PERSON": "#FF6B6B",  # 珊瑚红
            "ORGANIZATION": "#FFD166",  # 阳光黄
            "NATIONALITY": "#F08A5D",  # 哈密瓜色
            "RELIGION": "#F4A261",  # 沙棕色
            "PROFESSION": "#E76F51",  # 赭色
            "TITLE": "#F7B267",  # 芒果色
            "POLITICAL_PARTY": "#E63946",  # 帝王红
            # --- 地点与地理 (蓝色系) ---
            "LOCATION": "#4D96FF",  # 天蓝色
            "CELESTIAL_BODY": "#6BCB77",  # 海绿色
            "WEATHER": "#A2D2FF",  # 淡蓝色
            # --- 时间与数字 (绿色/青色系) ---
            "DATE": "#84D2C5",  # 薄荷绿
            "TIME": "#A6D6D6",  # 淡青色
            "MONEY": "#34A853",  # 森林绿
            "PERCENTAGE": "#2A9D8F",  # 丛林绿
            "MEASUREMENT": "#264653",  # 木炭蓝
            "CURRENCY": "#60D394",  # 淡绿色
            "STOCK_SYMBOL": "#1E8449",  # 深绿色
            # --- 创意与媒体 (紫色/粉色系) ---
            "PRODUCT": "#AB47BC",  # 紫色
            "ARTWORK": "#EC407A",  # 粉色
            "BOOK": "#7E57C2",  # 深紫色
            "MOVIE": "#9C27B0",  # 品红色
            "TV_SHOW": "#BA68C8",  # 淡紫色
            "MUSIC_GENRE": "#D81B60",  # 蔓越莓红
            "INSTRUMENT": "#8E24AA",  # 暗品红
            # --- 科技与科学 (青色/蓝色/灰色系) ---
            "TECHNOLOGY": "#00BCD4",  # 青色
            "SOFTWARE": "#00ACC1",  # 暗青色
            "HARDWARE": "#26C6DA",  # 淡青色
            "PROGRAMMING_LANGUAGE": "#4DD0E1",  # 亮青色
            "SCIENTIFIC_THEORY": "#03A9F4",  # 淡蓝色
            "CHEMICAL": "#B2EBF2",  # 粉末蓝
            "ACADEMIC_SUBJECT": "#81D4FA",  # 婴儿蓝
            # --- 自然与生物 (绿色/棕色系) ---
            "ANIMAL": "#A1887F",  # 胡桃棕
            "PLANT": "#4CAF50",  # 绿色
            "DISEASE": "#C62828",  # 暗红色
            "MEDICATION": "#EF5350",  # 淡红色
            "MEDICAL_PROCEDURE": "#E57373",  # 鲑鱼粉
            "FOOD": "#FF8A65",  # 淡橙色
            "DRINK": "#FFB74D",  # 暗黄色
            # --- 抽象与其他 (灰色/杂色系) ---
            "EVENT": "#FFC107",  # 琥珀色
            "LANGUAGE": "#795548",  # 棕色
            "AWARD": "#FFEB3B",  # 黄色
            "LAW": "#607D8B",  # 蓝灰色
            "CRIME": "#455A64",  # 暗蓝灰色
            "MATERIAL": "#BDBDBD",  # 淡灰色
            "COLOR": "#9E9E9E",  # 中灰色
            "SHAPE": "#757575",  # 暗灰色
            "VEHICLE": "#6D4C41",  # 深棕色
            "SPORT": "#009688",  # 蓝绿色
            "FILE_TYPE": "#8D6E63",  # fawn
            # --- 默认颜色 ---
            "DEFAULT": "#B0BEC5",  # 石板灰
        }

    # 2. 遍历 networkx 图的节点，并动态添加样式
    for node_id, node_attrs in g.nodes(data=True):
        label = node_attrs.get("entity_name", str(node_id))

        # 根据 entity_type 设置颜色
        entity_type = node_attrs.get("entity_type", "DEFAULT")
        color = entity_color_map.get(entity_type, entity_color_map["DEFAULT"])

        # 根据 pagerank 设置大小 (基础大小 + pagerank * 缩放因子)
        pagerank = node_attrs.get("pagerank", 0.0)
        # 调整缩放因子以获得最佳视觉效果，例如 1000
        size = 15 + (pagerank * 1000)

        net.add_node(
            node_id,
            label=label,
            title=node_attrs.get("description", label).replace("<SEP>", "\n"),
            color=color,
            size=size,
        )

    # 3. 遍历 networkx 图的边，并动态添加样式
    for source, target, edge_attrs in g.edges(data=True):
        edge_title = edge_attrs.get("description", "").replace("<SEP>", "\n")

        original_weight = float(edge_attrs.get("weight", 1.0))
        safe_weight = max(0, original_weight)
        scaled_width = 0.5 + math.log1p(safe_weight)

        net.add_edge(source, target, title=edge_title, width=scaled_width)

    # 4. 添加交互式UI控件
    # net.show_buttons(filter_=["nodes", "edges", "physics"])

    # 5. 生成 HTML 内容

    html_content = net.generate_html()
    # --- 解决浏览器缓存问题的关键 ---
    # 在 HTML 头部注入禁止缓存的 meta 标签
    cache_busting_tags = """
            <meta http-equiv="Cache-Control" content="no-cache, no-store, must-revalidate" />
            <meta http-equiv="Pragma" content="no-cache" />
            <meta http-equiv="Expires" content="0" />
            """
    html_content = html_content.replace("</head>", f"{cache_busting_tags}</head>")
    # ---------------------------------
    return html_content



import math
from typing import List, Optional, Any

from pydantic import BaseModel, Field

from duowen_agent.agents.component.base import BaseComponent
from duowen_agent.error import ObserverException
from duowen_agent.llm.chat_model import BaseAIChat
from duowen_agent.llm.tokenizer import tokenizer
from duowen_agent.prompt.prompt_build import GeneralPromptBuilder
from duowen_agent.utils.core_utils import json_observation, stream_to_string
from duowen_agent.utils.core_utils import retrying
from duowen_agent.utils.string_template import StringTemplate


class Keywords(BaseModel):
    keywords: List[str] = Field(description="重要的关键词/短语")


class ChunkKeywordsExtract(BaseComponent):

    def __init__(self, llm_instance: BaseAIChat, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.llm_instance = llm_instance
        self.kwargs = kwargs

    @staticmethod
    def build_prompt():
        return GeneralPromptBuilder(
            instruction="""角色：你是一个文本分析器
任务：提取给定文本内容中最重要的关键词/短语""",
            output_format=Keywords,
            sample="""```json
{
  "questions": [
    "关键词/短语1", 
    "关键词/短语2",
    "关键词/短语3",
    ...
  ]
}
```""",
            note=StringTemplate(
                """- 对文本内容进行总结，并给出前{topn}个最重要的关键词/短语
- 关键词/短语必须使用原文的语言"""
            ),
        )

    async def arun(self, *args, **kwargs) -> Any:
        raise NotImplementedError

    def run(
        self, chunk: str, top_n=3, title: str = None, summary: str = None
    ) -> Keywords:
        _question = ""
        if title:
            _question += f"- 文章标题: {title}\n\n"
        if summary:
            _question += f"- 文章摘要: {summary}\n\n"
        _question += f"- 文章片段内容: {chunk}\n\n"
        _question += f"---------------------\n\n"
        _question += "文章标题和文章摘要只是用来帮助你理解文章片段内容，请提取关于文章片段内容最重要的{topn}个关键词/短语"
        _prompt = self.build_prompt().get_instruction(
            user_input=_question, temp_vars={"topn": top_n}
        )
        _res = stream_to_string(self.llm_instance.chat_for_stream(_prompt))
        _res = json_observation(_res, Keywords)
        return _res


class Questions(BaseModel):
    questions: List[str] = Field(
        ..., description="propose questions about text content"
    )


class ChunkQuestionsExtract(BaseComponent):

    def __init__(self, llm_instance: BaseAIChat, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.llm_instance = llm_instance
        self.kwargs = kwargs

    @staticmethod
    def build_prompt():
        return GeneralPromptBuilder(
            instruction=StringTemplate(
                """Role: 你是一个文本分析器
Task: 针对给定的文本内容，提出{topn}个相关问题

## 约束条件（重要！）
- 必须基于文本内容直接生成
- 问题应具有明确答案指向性
- 需覆盖文本的不同方面
- 禁止生成假设性、重复或相似问题

## 处理流程
1. 【文本解析】分段处理内容，识别关键实体和核心概念
2. 【问题生成】基于信息密度选择最佳提问点
3. 【质量检查】确保：
   - 问题答案可在原文中找到依据
   - 标签与问题内容强相关
   - 无格式错误
"""
            ),
            output_format=Questions,
            sample="""```json
{
  "questions": [
    "问题1", 
    "问题2",
    "问题3",
    ...
  ]
}
```""",
            note=StringTemplate(
                """- 理解并总结文本内容后，提出最重要的{topn}个核心问题
- 问题必须聚焦文本的关键信息点
- 问题间应形成逻辑递进关系
- 须使用文本原语言进行提问
- 提出的问题不要和材料本身相关，例如禁止出现作者、章节、目录等相关问题
- 提出的问题中不得出现 ' 文章中 / 文章片段中 / 报告中 ' 等任何引用性表述
- """
            ),
        )

    # - 禁止出现"这些"、"它"、"该"、"本文"、"报告中"、"文章中"、"文献中"、"表格中"、"文章片段中"等类似指代性表述

    async def arun(self, *args, **kwargs) -> Any:
        raise NotImplementedError

    def run(
        self, chunk: str, top_n=3, title: str = None, summary: str = None
    ) -> Questions:
        _question = ""
        if title:
            _question += f"- 文章标题: {title}\n\n"
        if summary:
            _question += f"- 文章摘要: {summary}\n\n"
        _question += f"- 文章片段内容: {chunk}\n\n"
        _question += f"---------------------\n\n"
        _question += "文章标题和文章摘要只是用来帮助你理解文章片段内容，请提取关于文章片段内容最重要的{topn}个核心问题\n\n"
        _prompt = self.build_prompt().get_instruction(
            user_input=_question, temp_vars={"topn": top_n}
        )
        _res = stream_to_string(self.llm_instance.chat_for_stream(_prompt))
        _res = json_observation(_res, Questions)
        return _res


class TitleSummary(BaseModel):
    title: str = Field(description="一个精准的标题（不超过15字）")
    summary: str = Field(description="一段凝练的摘要（80-120字）")


class SummaryExtract(BaseComponent):
    def __init__(self, llm_instance: BaseAIChat, **kwargs):
        super().__init__(**kwargs)
        self.llm_instance = llm_instance
        self.kwargs = kwargs

    @staticmethod
    def build_prompt():
        return GeneralPromptBuilder(
            instruction="""请基于提供的文章片段，生成一个精准的标题（不超过15字）和一段凝练的摘要（80-120字），要求：
1. 标题需突出核心观点/发现/冲突
2. 摘要需包含三个关键要素：
   - 研究/事件的背景/目标
   - 使用的方法/核心论据
   - 主要结论/社会影响/争议点
3. 遵循学术/新闻规范（二选一）
""",
            output_format=TitleSummary,
            sample='''"""```json
{
  "title": "一个精准的标题（不超过15字）",
  "summary": "一段凝练的摘要（80-120字）"
}
```"""''',
            note="""- 若原文含数据/专有名词，必须保留关键数值和术语
- 不添加解释性内容，不采用"本文""本研究"等主观表述""",
        )

    def run(self, content: str) -> TitleSummary:
        _content = tokenizer.truncate_chat(
            content, int(self.llm_instance.token_limit / 2)
        )
        _prompt = self.build_prompt().get_instruction(
            user_input=f"文章片段:\n\n{_content}"
        )
        _res = stream_to_string(self.llm_instance.chat_for_stream(_prompt))
        _res = json_observation(_res, TitleSummary)
        return _res

    async def arun(self, *args, **kwargs) -> Any:
        raise NotImplementedError


class ChunkExtendList(BaseModel):
    index: int
    data: Optional[List[str]]


class ChunkExtendSchema(BaseModel):
    title: Optional[str]
    summary: Optional[str]
    chunk_keywords: Optional[List[ChunkExtendList]]
    chunk_questions: Optional[List[ChunkExtendList]]


class ChunkExtend(BaseComponent):

    def __init__(self, llm_instance: BaseAIChat, **kwargs):
        super().__init__(**kwargs)
        self.llm_instance = llm_instance
        self.kwargs = kwargs

    async def arun(self, *args, **kwargs) -> Any:
        raise NotImplementedError

    def run(
        self,
        content: str,
        chunks: List[str],
        keywords_dividend: int = 120,
        questions_dividend: int = 180,
        title: str = None,
        is_summary: bool = True,
    ) -> ChunkExtendSchema:
        _title = title
        _summary = None
        if is_summary:
            try:
                _res = retrying(
                    func=SummaryExtract(self.llm_instance).run,
                    func_params=dict(content=content),
                )
                if not _title:
                    _title = _res.title
                _summary = _res.summary
            except ObserverException:
                pass

        _chunk_keywords = []
        if keywords_dividend:
            for _index, _chunk in enumerate(chunks):
                _keywords_topn = math.floor(len(_chunk) / keywords_dividend)
                if _keywords_topn >= 1:
                    try:
                        _res = retrying(
                            func=ChunkKeywordsExtract(self.llm_instance).run,
                            func_params=dict(
                                chunk=_chunk,
                                top_n=_keywords_topn,
                                title=_title,
                                summary=_summary,
                            ),
                        )
                        _chunk_keywords.append(
                            {"index": _index, "data": _res.keywords[:_keywords_topn]}
                        )
                    except ObserverException:
                        _chunk_keywords.append({"index": _index, "data": []})
                else:
                    _chunk_keywords.append({"index": _index, "data": []})

        _chunk_questions = []
        if questions_dividend:
            for _index, _chunk in enumerate(chunks):
                _questions_topn = math.floor(len(_chunk) / questions_dividend)
                if _questions_topn >= 1:
                    try:
                        _res = retrying(
                            func=ChunkQuestionsExtract(self.llm_instance).run,
                            func_params=dict(
                                chunk=_chunk,
                                top_n=_questions_topn,
                                title=_title,
                                summary=_summary,
                            ),
                        )
                        _chunk_questions.append(
                            {"index": _index, "data": _res.questions[:_questions_topn]}
                        )
                    except ObserverException:
                        _chunk_questions.append({"index": _index, "data": []})
                else:
                    _chunk_questions.append({"index": _index, "data": []})

        return ChunkExtendSchema(
            title=_title,
            summary=_summary,
            chunk_keywords=_chunk_keywords,
            chunk_questions=_chunk_questions,
        )



import html
import io
from typing import Union

import pptx
from markdownify import markdownify
from pptx.enum.shapes import MSO_SHAPE_TYPE


class PptxExtractor:
    def convert(self, local_path: Union[str, io.BytesIO]) -> str:
        md_content = ""

        presentation = pptx.Presentation(local_path)
        slide_num = 0
        for slide in presentation.slides:
            slide_num += 1
            md_content += f"\n\n<!-- Slide number: {slide_num} -->\n"
            title = slide.shapes.title
            for shape in slide.shapes:
                # Tables
                if self._is_table(shape):
                    html_table = "<html><body><table>"
                    first_row = True
                    for row in shape.table.rows:
                        html_table += "<tr>"
                        for cell in row.cells:
                            if first_row:
                                html_table += "<th>" + html.escape(cell.text) + "</th>"
                            else:
                                html_table += "<td>" + html.escape(cell.text) + "</td>"
                        html_table += "</tr>"
                        first_row = False
                    html_table += "</table></body></html>"
                    md_content += (
                            "\n" + markdownify(html_table).strip() + "\n"
                    )

                # Charts
                if shape.has_chart:
                    md_content += self._convert_chart_to_markdown(shape.chart)

                # Text areas
                elif shape.has_text_frame:
                    if shape == title:
                        md_content += "# " + shape.text.lstrip() + "\n"
                    else:
                        md_content += shape.text + "\n"

            md_content = md_content.strip()

        return md_content

    @classmethod
    def _is_table(cls, shape):
        return shape.shape_type == MSO_SHAPE_TYPE.TABLE

    @classmethod
    def _convert_chart_to_markdown(cls, chart):
        md = "\n\n### Chart"
        if chart.has_title:
            md += f": {chart.chart_title.text_frame.text}"
        md += "\n\n"
        data = []
        category_names = [c.label for c in chart.plots[0].categories]
        series_names = [s.name for s in chart.series]
        data.append(["Category"] + series_names)

        for idx, category in enumerate(category_names):
            row = [category]
            for series in chart.series:
                row.append(series.values[idx])
            data.append(row)

        markdown_table = []
        for row in data:
            markdown_table.append("| " + " | ".join(map(str, row)) + " |")
        header = markdown_table[0]
        separator = "|" + "|".join(["---"] * len(data[0])) + "|"
        return md + "\n".join([header, separator] + markdown_table[1:])



from abc import ABC, abstractmethod
from typing import List, Optional

import numpy as np
from duowen_agent.llm import OpenAIEmbedding
from duowen_agent.rag.nlp import LexSynth

from ..models import Document, SearchResult


class BaseVector(ABC):
    def __init__(self, lex_synth: LexSynth, llm_embeddings_instance: OpenAIEmbedding):
        self.lex_synth = lex_synth
        self.llm_embeddings_instance = llm_embeddings_instance

    @abstractmethod
    def get_backend_type(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def semantic_search(
        self,
        query: str,
        top_k: int = 10,
        score_threshold: Optional[float] = 0.0,
        query_embedding: np.ndarray | List[float] = None,
        **kwargs
    ) -> list[SearchResult]:
        raise NotImplementedError

    @abstractmethod
    def full_text_search(
        self,
        query: str,
        top_k: int = 10,
        score_threshold: Optional[float] = 0.0,
        **kwargs
    ) -> list[SearchResult]:
        raise NotImplementedError

    @abstractmethod
    def add_document(self, document: Document, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def batch_add_document(self, documents: list[Document], batch_num: int, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def get_documents_by_ids(self, ids: list[str], **kwargs) -> list[Document]:
        raise NotImplementedError

    @abstractmethod
    def delete_documents_by_ids(self, ids: list[str], **kwargs):
        raise NotImplementedError

    def hybrid_search(
        self,
        query: str,
        top_k: int = 10,
        score_threshold: Optional[float] = 0.0,
        query_embedding: np.ndarray | List[float] = None,
        vector_weight: float = 0.7,
        keyword_weight: float = 0.3,
        **kwargs
    ) -> list[SearchResult]:

        if not query_embedding:
            query_embedding = self.llm_embeddings_instance.get_embedding(query)[0]

        full_text_data = self.full_text_search(query, top_k=top_k, **kwargs)
        semantic_data = self.semantic_search(
            query, top_k=top_k, query_embedding=query_embedding, **kwargs
        )

        _data: List[Document] = list(
            set([i.result for i in full_text_data] + [i.result for i in semantic_data])
        )

        hybrid_data = self.lex_synth.hybrid_similarity(
            question=query,
            question_vector=query_embedding,
            docs_vector=[i.vector for i in _data],
            docs_sm=[
                (
                    i.page_content_split
                    if i.page_content_split
                    else self.lex_synth.content_cut(i.page_content)
                )
                for i in _data
            ],
            tkweight=keyword_weight,
            vtweight=vector_weight,
        )

        hybrid_data_index = [(index, score) for index, score in enumerate(hybrid_data)]

        hybrid_data_index = sorted(hybrid_data_index, key=lambda x: x[1], reverse=True)[
            :top_k
        ]

        return [
            SearchResult(result=_data[i[0]], rerank_similarity_score=i[1])
            for i in hybrid_data_index
            if i[1] >= score_threshold
        ]



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
# SPDX-License-Identifier: MIT

"""
Built-in questions for Deer.
"""

# English built-in questions
BUILT_IN_QUESTIONS = [
    "What factors are influencing AI adoption in healthcare?",
    "How does quantum computing impact cryptography?",
    "What are the latest developments in renewable energy technology?",
    "How is climate change affecting global agriculture?",
    "What are the ethical implications of artificial intelligence?",
    "What are the current trends in cybersecurity?",
    "How is blockchain technology being used outside of cryptocurrency?",
    "What advances have been made in natural language processing?",
    "How is machine learning transforming the financial industry?",
    "What are the environmental impacts of electric vehicles?",
]

# Chinese built-in questions
BUILT_IN_QUESTIONS_ZH_CN = [
    "人工智能在医疗保健领域的应用有哪些因素影响?",
    "量子计算如何影响密码学?",
    "可再生能源技术的最新发展是什么?",
    "气候变化如何影响全球农业?",
    "人工智能的伦理影响是什么?",
    "网络安全的当前趋势是什么?",
    "区块链技术在加密货币之外如何应用?",
    "自然语言处理领域有哪些进展?",
    "机器学习如何改变金融行业?",
    "电动汽车对环境有什么影响?",
]



from typing import Literal, Optional, List

from pydantic import BaseModel

from duowen_agent.agents.react import ReactObservation, ReactAction, ReactResult


class MsgInfo(BaseModel):
    content: str


class HumanFeedbackInfo(BaseModel):
    type: Literal["str"] = "str"
    content: str


class PlanStep(BaseModel):
    title: str
    description: Optional[str]
    status: Literal["success", "running", "failed", "waiting"]


class PlanInfo(BaseModel):
    title: str
    steps: List[PlanStep]


class ReactStartInfo(BaseModel):
    content: str
    node_id: str


class ReactObservationInfo(ReactObservation):
    node_id: str


class ReactActionInfo(ReactAction):
    node_id: str


class ReactResultInfo(ReactResult):
    node_id: str


class ReactEndInfo(BaseModel):
    content: str
    node_id: str


class ResultInfo(BaseModel):
    type: Literal["msg", "markdown"]
    file_name: Optional[str] = None
    content: str



"""markdown 表格解析"""

import re

import mistune
from bs4 import BeautifulSoup

# 用户提供的正则表达式模式
TABLE_PATTERN = (
    r"^ {0,3}\|(?P<table_head>.+)\|[ \t]*\n"
    r" {0,3}\|(?P<table_align> *[-:]+[-| :]*)\|[ \t]*\n"
    r"(?P<table_body>(?: {0,3}\|.*\|[ \t]*(?:\n|$))*)\n*"
)
NP_TABLE_PATTERN = (
    r"^ {0,3}(?P<nptable_head>\S.*\|.*)\n"
    r" {0,3}(?P<nptable_align>[-:]+ *\|[-| :]*)\n"
    r"(?P<nptable_body>(?:.*\|.*(?:\n|$))*)\n*"
)


def markdown_table_to_kv_list(text):
    # 查找所有表格并记录位置
    tables = []
    seen = set()

    # 匹配标准表格
    for match in re.finditer(TABLE_PATTERN, text, flags=re.MULTILINE):
        start, end = match.span()
        if (start, end) not in seen:
            table_text = match.group(0)
            tables.append((start, end, table_text))
            seen.add((start, end))

    # 匹配非标准表格
    for match in re.finditer(NP_TABLE_PATTERN, text, flags=re.MULTILINE):
        start, end = match.span()
        if (start, end) not in seen:
            table_text = match.group(0)
            tables.append((start, end, table_text))
            seen.add((start, end))

    if not tables:
        return text

    # print(tables)

    # 按起始位置逆序排序
    tables_sorted = sorted(tables, key=lambda x: -x[0])

    # 初始化Markdown解析器（确保已安装mistune并启用表格插件）
    md = mistune.create_markdown(escape=True, plugins=["table"])

    # 处理每个表格并构建替换段
    segments = []
    last_end = len(text)
    for start, end, table_text in tables_sorted:
        # 转换为HTML
        html = md(table_text.strip())

        if not (
            html.strip().startswith("<table>") and not html.strip().endswith("</table>")
        ):
            return text

        # 转换为JSONL
        jsonl_str = html_table_to_kv_list(html) + "\n\n"

        # 记录替换段
        if jsonl_str.strip():
            segments.append((start, end, jsonl_str))

    # 构建替换后的文本（从后向前替换）
    new_text = []
    last_pos = len(text)
    for start, end, jsonl_str in sorted(segments, key=lambda x: -x[0]):
        new_text.append(text[end:last_pos])
        new_text.append(jsonl_str)
        last_pos = start
    new_text.append(text[:last_pos])
    new_text.reverse()

    return "".join(new_text)


def html_table_to_kv_list(html):
    soup = BeautifulSoup(html, "html.parser")
    table = soup.find("table")
    if not table:
        return ""

    # 提取表头（处理合并单元格）
    headers = []
    header_row = None

    # 尝试从thead或第一个tr获取表头
    if thead := table.find("thead"):
        header_row = thead.find("tr")
    elif first_tr := table.find("tr"):
        header_row = first_tr

    if header_row:
        for cell in header_row.find_all(["th", "td"]):
            # 处理colspan合并
            colspan = int(cell.get("colspan", 1))
            cell_text = cell.get_text(strip=True)
            headers.extend([cell_text] * colspan)

    # 提取数据行（处理合并单元格）
    rows = []
    tbody = table.find("tbody") or table

    for row in tbody.find_all("tr"):
        # 跳过表头行
        if row == header_row:
            continue

        # 处理单元格合并
        cells = []
        for cell in row.find_all(["td", "th"]):
            # 处理colspan合并
            colspan = int(cell.get("colspan", 1))
            cell_text = cell.get_text(strip=True)
            cells.extend([cell_text] * colspan)

        # 生成键值序列
        if len(cells) == len(headers):
            pairs = [f"{key}:{value}" for key, value in zip(headers, cells)]
            rows.append(";".join(pairs))

    return "\n".join(rows)



from datetime import datetime
from typing import Literal, List

from duowen_agent.agents.component.base import BaseComponent
from duowen_agent.llm import MessagesSet
from duowen_agent.utils.core_utils import stream_to_string, remove_think
from duowen_agent.utils.string_template import StringTemplate
from .classifiers import ClassifiersOne


class PromptEnhancer(BaseComponent):

    def __init__(self, llm, lang: Literal["en", "cn"] = "cn", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.llm = llm
        self.lang = lang
        self._system_prompt = {
            "en": StringTemplate(
                """---
CURRENT_TIME: {{ CURRENT_TIME }}
---

You are an expert prompt engineer. Your task is to enhance user prompts to make them more effective, specific, and likely to produce high-quality results from AI systems.

# Your Role
- Analyze the original prompt for clarity, specificity, and completeness
- Enhance the prompt by adding relevant details, context, and structure
- Make the prompt more actionable and results-oriented
- Preserve the user's original intent while improving effectiveness

{% if report_style == "academic" %}
# Enhancement Guidelines for Academic Style
1. **Add methodological rigor**: Include research methodology, scope, and analytical framework
2. **Specify academic structure**: Organize with clear thesis, literature review, analysis, and conclusions
3. **Clarify scholarly expectations**: Specify citation requirements, evidence standards, and academic tone
4. **Add theoretical context**: Include relevant theoretical frameworks and disciplinary perspectives
5. **Ensure precision**: Use precise terminology and avoid ambiguous language
6. **Include limitations**: Acknowledge scope limitations and potential biases
{% elif report_style == "popular_science" %}
# Enhancement Guidelines for Popular Science Style
1. **Add accessibility**: Transform technical concepts into relatable analogies and examples
2. **Improve narrative structure**: Organize as an engaging story with clear beginning, middle, and end
3. **Clarify audience expectations**: Specify general audience level and engagement goals
4. **Add human context**: Include real-world applications and human interest elements
5. **Make it compelling**: Ensure the prompt guides toward fascinating and wonder-inspiring content
6. **Include visual elements**: Suggest use of metaphors and descriptive language for complex concepts
{% elif report_style == "news" %}
# Enhancement Guidelines for News Style
1. **Add journalistic rigor**: Include fact-checking requirements, source verification, and objectivity standards
2. **Improve news structure**: Organize with inverted pyramid structure (most important information first)
3. **Clarify reporting expectations**: Specify timeliness, accuracy, and balanced perspective requirements
4. **Add contextual background**: Include relevant background information and broader implications
5. **Make it newsworthy**: Ensure the prompt focuses on current relevance and public interest
6. **Include attribution**: Specify source requirements and quote standards
{% elif report_style == "social_media" %}
# Enhancement Guidelines for Social Media Style
1. **Add engagement focus**: Include attention-grabbing elements, hooks, and shareability factors
2. **Improve platform structure**: Organize for specific platform requirements (character limits, hashtags, etc.)
3. **Clarify audience expectations**: Specify target demographic and engagement goals
4. **Add viral elements**: Include trending topics, relatable content, and interactive elements
5. **Make it shareable**: Ensure the prompt guides toward content that encourages sharing and discussion
6. **Include visual considerations**: Suggest emoji usage, formatting, and visual appeal elements
{% else %}
# General Enhancement Guidelines
1. **Add specificity**: Include relevant details, scope, and constraints
2. **Improve structure**: Organize the request logically with clear sections if needed
3. **Clarify expectations**: Specify desired output format, length, or style
4. **Add context**: Include background information that would help generate better results
5. **Make it actionable**: Ensure the prompt guides toward concrete, useful outputs
{% endif %}

# Output Requirements
- Output ONLY the enhanced prompt
- Do NOT include any explanations, comments, or meta-text
- Do NOT use phrases like "Enhanced Prompt:" or "Here's the enhanced version:"
- The output should be ready to use directly as a prompt

{% if report_style == "academic" %}
# Academic Style Examples

**Original**: "Write about AI"
**Enhanced**: "Conduct a comprehensive academic analysis of artificial intelligence applications across three key sectors: healthcare, education, and business. Employ a systematic literature review methodology to examine peer-reviewed sources from the past five years. Structure your analysis with: (1) theoretical framework defining AI and its taxonomies, (2) sector-specific case studies with quantitative performance metrics, (3) critical evaluation of implementation challenges and ethical considerations, (4) comparative analysis across sectors, and (5) evidence-based recommendations for future research directions. Maintain academic rigor with proper citations, acknowledge methodological limitations, and present findings with appropriate hedging language. Target length: 3000-4000 words with APA formatting."

**Original**: "Explain climate change"
**Enhanced**: "Provide a rigorous academic examination of anthropogenic climate change, synthesizing current scientific consensus and recent research developments. Structure your analysis as follows: (1) theoretical foundations of greenhouse effect and radiative forcing mechanisms, (2) systematic review of empirical evidence from paleoclimatic, observational, and modeling studies, (3) critical analysis of attribution studies linking human activities to observed warming, (4) evaluation of climate sensitivity estimates and uncertainty ranges, (5) assessment of projected impacts under different emission scenarios, and (6) discussion of research gaps and methodological limitations. Include quantitative data, statistical significance levels, and confidence intervals where appropriate. Cite peer-reviewed sources extensively and maintain objective, third-person academic voice throughout."

{% elif report_style == "popular_science" %}
# Popular Science Style Examples

**Original**: "Write about AI"
**Enhanced**: "Tell the fascinating story of how artificial intelligence is quietly revolutionizing our daily lives in ways most people never realize. Take readers on an engaging journey through three surprising realms: the hospital where AI helps doctors spot diseases faster than ever before, the classroom where intelligent tutors adapt to each student's learning style, and the boardroom where algorithms are making million-dollar decisions. Use vivid analogies (like comparing neural networks to how our brains work) and real-world examples that readers can relate to. Include 'wow factor' moments that showcase AI's incredible capabilities, but also honest discussions about current limitations. Write with infectious enthusiasm while maintaining scientific accuracy, and conclude with exciting possibilities that await us in the near future. Aim for 1500-2000 words that feel like a captivating conversation with a brilliant friend."

**Original**: "Explain climate change"
**Enhanced**: "Craft a compelling narrative that transforms the complex science of climate change into an accessible and engaging story for curious readers. Begin with a relatable scenario (like why your hometown weather feels different than when you were a kid) and use this as a gateway to explore the fascinating science behind our changing planet. Employ vivid analogies - compare Earth's atmosphere to a blanket, greenhouse gases to invisible heat-trapping molecules, and climate feedback loops to a snowball rolling downhill. Include surprising facts and 'aha moments' that will make readers think differently about the world around them. Weave in human stories of scientists making discoveries, communities adapting to change, and innovative solutions being developed. Balance the serious implications with hope and actionable insights, concluding with empowering steps readers can take. Write with wonder and curiosity, making complex concepts feel approachable and personally relevant."

{% elif report_style == "news" %}
# News Style Examples

**Original**: "Write about AI"
**Enhanced**: "Report on the current state and immediate impact of artificial intelligence across three critical sectors: healthcare, education, and business. Lead with the most newsworthy developments and recent breakthroughs that are affecting people today. Structure using inverted pyramid format: start with key findings and immediate implications, then provide essential background context, followed by detailed analysis and expert perspectives. Include specific, verifiable data points, recent statistics, and quotes from credible sources including industry leaders, researchers, and affected stakeholders. Address both benefits and concerns with balanced reporting, fact-check all claims, and provide proper attribution for all information. Focus on timeliness and relevance to current events, highlighting what's happening now and what readers need to know. Maintain journalistic objectivity while making the significance clear to a general news audience. Target 800-1200 words following AP style guidelines."

**Original**: "Explain climate change"
**Enhanced**: "Provide comprehensive news coverage of climate change that explains the current scientific understanding and immediate implications for readers. Lead with the most recent and significant developments in climate science, policy, or impacts that are making headlines today. Structure the report with: breaking developments first, essential background for understanding the issue, current scientific consensus with specific data and timeframes, real-world impacts already being observed, policy responses and debates, and what experts say comes next. Include quotes from credible climate scientists, policy makers, and affected communities. Present information objectively while clearly communicating the scientific consensus, fact-check all claims, and provide proper source attribution. Address common misconceptions with factual corrections. Focus on what's happening now, why it matters to readers, and what they can expect in the near future. Follow journalistic standards for accuracy, balance, and timeliness."

{% elif report_style == "social_media" %}
# Social Media Style Examples

**Original**: "Write about AI"
**Enhanced**: "Create engaging social media content about AI that will stop the scroll and spark conversations! Start with an attention-grabbing hook like 'You won't believe what AI just did in hospitals this week 🤯' and structure as a compelling thread or post series. Include surprising facts, relatable examples (like AI helping doctors spot diseases or personalizing your Netflix recommendations), and interactive elements that encourage sharing and comments. Use strategic hashtags (#AI #Technology #Future), incorporate relevant emojis for visual appeal, and include questions that prompt audience engagement ('Have you noticed AI in your daily life? Drop examples below! 👇'). Make complex concepts digestible with bite-sized explanations, trending analogies, and shareable quotes. Include a clear call-to-action and optimize for the specific platform (Twitter threads, Instagram carousel, LinkedIn professional insights, or TikTok-style quick facts). Aim for high shareability with content that feels both informative and entertaining."

**Original**: "Explain climate change"
**Enhanced**: "Develop viral-worthy social media content that makes climate change accessible and shareable without being preachy. Open with a scroll-stopping hook like 'The weather app on your phone is telling a bigger story than you think 📱🌡️' and break down complex science into digestible, engaging chunks. Use relatable comparisons (Earth's fever, atmosphere as a blanket), trending formats (before/after visuals, myth-busting series, quick facts), and interactive elements (polls, questions, challenges). Include strategic hashtags (#ClimateChange #Science #Environment), eye-catching emojis, and shareable graphics or infographics. Address common questions and misconceptions with clear, factual responses. Create content that encourages positive action rather than climate anxiety, ending with empowering steps followers can take. Optimize for platform-specific features (Instagram Stories, TikTok trends, Twitter threads) and include calls-to-action that drive engagement and sharing."

{% else %}
# General Examples

**Original**: "Write about AI"
**Enhanced**: "Write a comprehensive 1000-word analysis of artificial intelligence's current applications in healthcare, education, and business. Include specific examples of AI tools being used in each sector, discuss both benefits and challenges, and provide insights into future trends. Structure the response with clear sections for each industry and conclude with key takeaways."

**Original**: "Explain climate change"
**Enhanced**: "Provide a detailed explanation of climate change suitable for a general audience. Cover the scientific mechanisms behind global warming, major causes including greenhouse gas emissions, observable effects we're seeing today, and projected future impacts. Include specific data and examples, and explain the difference between weather and climate. Organize the response with clear headings and conclude with actionable steps individuals can take."
{% endif %}""",
                template_format="jinja2",
            ),
            "cn": StringTemplate(
                """---
CURRENT_TIME: {{ CURRENT_TIME }}
---

你是一位专业的提示工程师。你的任务是增强用户的提示词，使其更有效、更具体，并更有可能从AI系统中获得高质量的结果。

# 你的角色
- 分析原始提示词的清晰度、具体性和完整性
- 通过添加相关细节、上下文和结构来增强提示词
- 使提示词更具可操作性和结果导向性
- 在提高有效性的同时，保留用户的原始意图

{% if report_style == "academic" %}
# 学术风格增强指南
1.  **增加方法论的严谨性**：包括研究方法、范围和分析框架
2.  **明确学术结构**：用清晰的论点、文献综述、分析和结论来组织内容
3.  **阐明学术期望**：指定引用要求、证据标准和学术语气
4.  **添加理论背景**：包括相关的理论框架和学科视角
5.  **确保精确性**：使用精确的术语，避免模糊语言
6.  **包含局限性**：承认范围限制和潜在偏见
{% elif report_style == "popular_science" %}
# 科普风格增强指南
1.  **增加可及性**：将技术概念转化为相关的类比和示例
2.  **改进叙事结构**：组织成一个有吸引力的故事，有清晰的开头、中间和结尾
3.  **阐明受众期望**：明确普通受众水平和参与目标
4.  **添加人文背景**：包括现实世界的应用和人文趣味元素
5.  **使其引人入胜**：确保提示词能引导出引人入胜和激发好奇心的内容
6.  **包含视觉元素**：建议使用隐喻和描述性语言来解释复杂概念
{% elif report_style == "news" %}
# 新闻风格增强指南
1.  **增加新闻严谨性**：包括事实核查要求、来源验证和客观性标准
2.  **改进新闻结构**：采用倒金字塔结构组织（最重要的信息放在最前面）
3.  **阐明报道期望**：指定时效性、准确性和平衡视角的要求
4.  **添加上下文背景**：包括相关的背景信息和更广泛的影响
5.  **突出新闻价值**：确保提示词侧重于当前相关性和公众兴趣
6.  **包含信息来源**：指定来源要求和引用标准
{% elif report_style == "social_media" %}
# 社交媒体风格增强指南
1.  **增加互动焦点**：包括吸引注意力的元素、钩子和可分享性因素
2.  **改进平台结构**：根据特定平台要求（字符限制、主题标签等）进行组织
3.  **阐明受众期望**：指定目标人群和互动目标
4.  **添加病毒式传播元素**：包括热门话题、相关内容（relatable content）和互动元素
5.  **使其易于分享**：确保提示词能引导出鼓励分享和讨论的内容
6.  **包含视觉考虑因素**：建议使用表情符号、格式和视觉吸引力元素
{% else %}
# 通用增强指南
1.  **增加具体性**：包括相关细节、范围和约束条件
2.  **改进结构**：如果需要，用清晰的部分有逻辑地组织请求
3.  **阐明期望**：指定期望的输出格式、长度或风格
4.  **添加上下文**：包括有助于产生更好结果的背景信息
5.  **使其可操作**：确保提示词能引导出具体、有用的输出
{% endif %}

# 输出要求
- 仅输出增强后的提示词
- 不要包含任何解释、评论或元文本
- 不要使用诸如“增强后的提示词：”或“这是增强版本：”之类的短语
- 输出应可直接用作提示词

{% if report_style == "academic" %}
# 学术风格示例

**原始提示**：“写关于人工智能”
**增强后**：“对人工智能在三个关键领域（医疗保健、教育和商业）的应用进行全面学术分析。采用系统性文献综述方法，审查过去五年内经同行评审的来源。按以下结构组织分析：(1) 定义AI及其分类的理论框架，(2) 包含定量绩效指标的特定行业案例研究，(3) 对实施挑战和伦理考虑的关键评估，(4) 跨行业比较分析，以及 (5) 基于证据的未来研究方向建议。保持学术严谨性，使用适当的引用，承认方法论局限性，并使用适当的谨慎语言呈现研究结果。目标长度：3000-4000词，采用APA格式。”

**原始提示**：“解释气候变化”
**增强后**：“对人为气候变化进行严格的学术考察，综合当前科学共识和近期研究进展。按以下结构组织分析：(1) 温室效应和辐射强迫机制的理论基础，(2) 对来自古气候、观测和模型研究的经验证据的系统性回顾，(3) 将人类活动与观测到的变暖联系起来的归因研究的关键分析，(4) 气候敏感性估计和不确定性范围的评估，(5) 不同排放情景下预期影响的评估，以及 (6) 研究差距和方法论局限性的讨论。酌情包含定量数据、统计显著性水平和置信区间。广泛引用同行评审来源，并始终保持客观的第三人称学术语气。”

{% elif report_style == "popular_science" %}
# 科普风格示例

**原始提示**：“写关于人工智能”
**增强后**：“讲述人工智能如何以大多数人从未意识到的方式悄然革新我们日常生活的迷人故事。带领读者踏上一段引人入胜的旅程，穿越三个令人惊讶的领域：AI帮助医生比以往更快发现疾病的医院，智能导师适应每个学生学习风格的教室，以及算法正在做出百万美元决策的董事会。使用生动的类比（比如将神经网络比作我们大脑的工作方式）和读者能够产生共鸣的真实示例。包含展示AI惊人能力的‘惊叹时刻’，但也诚实地讨论当前的局限性。以富有感染力的热情写作，同时保持科学准确性，并以不久的将来等待着我们的激动人心的可能性作为结尾。目标字数1500-2000，读起来像与一位才华横溢的朋友进行迷人的对话。”

**原始提示**：“解释气候变化”
**增强后**：“将一个复杂的气候变化科学转化为易于理解且引人入胜的故事，献给好奇的读者。从一个相关的场景开始（比如为什么你家乡的天气感觉和你小时候不一样了），并以此作为切入点，探索我们星球变化背后迷人的科学。使用生动的类比——将地球大气层比作毯子，温室气体比作看不见的吸热分子，气候反馈循环比作下坡的雪球。包含令人惊讶的事实和‘顿悟时刻’，让读者以不同的方式思考周围的世界。穿插科学家们做出发现、社区适应变化以及正在开发的创新解决方案的人文故事。在严肃的影响与希望和可行的见解之间取得平衡，最后给出读者可以采取的有效步骤。带着惊奇和好奇心写作，使复杂的概念变得平易近人且与个人相关。”

{% elif report_style == "news" %}
# 新闻风格示例

**原始提示**：“写关于人工智能”
**增强后**：“报道人工智能在三个关键领域（医疗保健、教育和商业）的现状和直接影响。以最具有新闻价值的发展和近期突破为首要内容，这些发展正在影响当今的人们。使用倒金字塔结构：从关键发现和直接影响开始，然后提供必要的背景信息，接着是详细的分析和专家观点。包含具体的、可验证的数据点、近期统计数据，以及来自行业领袖、研究人员和受影响利益相关者等可信来源的引述。以平衡的报道方式阐述益处和担忧，对所有声明进行事实核查，并为所有信息提供适当的来源说明。侧重于时效性和与当前事件的相关性，突出现在正在发生的事情以及读者需要了解的内容。在向普通新闻受众阐明其重要性的同时，保持新闻客观性。目标字数800-1200，遵循美联社（AP）风格指南。”

**原始提示**：“解释气候变化”
**增强后**：“提供关于气候变化的全面新闻报道，解释当前的科学理解以及对读者的直接影响。以气候科学、政策或影响方面最近和最重要的发展为首要内容，这些内容正在成为今日头条。按以下结构组织报告：首先是最新进展，然后是理解该问题所需的基本背景，接着是包含具体数据和时间范围的当前科学共识，已经观察到的现实世界影响，政策回应和辩论，以及专家对下一步的看法。引用可信的气候科学家、政策制定者和受影响社区的言论。客观地呈现信息，同时清晰地传达科学共识，对所有声明进行事实核查，并提供适当的信息来源说明。用事实纠正常见的误解。侧重于现在正在发生的事情、它与读者的关系以及他们在不久的将来可以预期的事情。遵循准确性、平衡性和时效性的新闻标准。”

{% elif report_style == "social_media" %}
# 社交媒体风格示例

**原始提示**：“写关于人工智能”
**增强后**：“创建关于人工智能的吸引人的社交媒体内容，旨在让用户停止滚动并引发对话！以一个吸引眼球的钩子开头，例如‘你绝对不敢相信AI这周在医院做了什么 🤯’，并将其组织成一个引人入胜的推文串或帖子系列。包含令人惊讶的事实、相关的例子（比如AI帮助医生发现疾病或个性化你的Netflix推荐），以及鼓励分享和评论的互动元素。使用策略性主题标签（#AI #技术 #未来），加入相关表情符号以增强视觉吸引力，并包含促使受众互动的问题（‘你在日常生活中注意到AI了吗？在下面留言分享例子！👇’）。通过简化的解释、流行的类比和可分享的引述，使复杂概念易于理解。包含清晰的行动号召（call-to-action），并针对特定平台进行优化（推特串、Instagram轮播图、LinkedIn专业见解或TikTok风格的快知识）。力求高分享度，使内容既 informative 又有趣。”

**原始提示**：“解释气候变化”
**增强后**：“开发具有病毒式传播潜力的社交媒体内容，使气候变化变得易于理解和分享，同时又不说教。以一个能让人停止滚动的钩子开头，例如‘你手机上的天气应用程序正在讲述一个比你想象的更大的故事 📱🌡️’，并将复杂的科学分解成易于消化、引人入胜的小块。使用相关的比较（地球发烧了，大气层像毯子），流行的格式（前后对比视觉图、辟谣系列、快知识），和互动元素（投票、问题、挑战）。包含策略性主题标签（#气候变化 #科学 #环境），吸引眼球的表情符号，以及可分享的图形或信息图。用清晰、事实性的回应解答常见问题和误解。创建鼓励积极行动而非气候焦虑的内容，最后给出关注者可以采取的赋能步骤。针对平台特定功能（Instagram Stories, TikTok trends, Twitter threads）进行优化，并包含能推动互动和分享的行动号召。”

{% else %}
# 通用示例

**原始提示**：“写关于人工智能”
**增强后**：“撰写一篇1000字的综合分析，探讨人工智能目前在医疗保健、教育和商业领域的应用。包括每个领域正在使用的AI工具的具体示例，讨论益处和挑战，并提供对未来趋势的见解。用清晰的章节结构组织每个行业的回应，并以关键要点作为结论。”

**原始提示**：“解释气候变化”
**增强后**：“为普通受众提供关于气候变化的详细解释。涵盖全球变暖背后的科学机制、包括温室气体排放在内的主要原因、我们今天观察到的明显影响以及预测的未来影响。包含具体数据和示例，并解释天气和气候之间的区别。用清晰的标题组织回应，并以个人可以采取的行动步骤作为结尾。”
{% endif %}
""",
                template_format="jinja2",
            ),
        }

        self.classify = {
            "academic": "涉及学术研究、专业理论、学科知识",
            "popular_science": "科学知识普及或生活科技解释",
            "news": "时效性事件、官方通报或公共事务",
            "social_media": "网络流行话题/社交平台内容",
            "general": "日常咨询/闲聊/无明确领域倾向",
        }

    def run(
        self,
        question: str,
        report_style: List[
            Literal["academic", "popular_science", "news", "social_media", "general"]
        ] = None,
    ) -> str:
        if report_style and len(report_style) == 1:
            _report_style = report_style[0]
        elif report_style and len(report_style) > 1:
            _report_style = ClassifiersOne(self.llm).run(
                question, {i: self.classify[i] for i in report_style}
            )
        else:
            _report_style = ClassifiersOne(self.llm).run(question, self.classify)
        _prompt = MessagesSet().add_system(
            self._system_prompt[self.lang].format(
                report_style=_report_style,
                CURRENT_TIME=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            )
        )
        _prompt.add_user(
            f"用户问题：{question} \n\n请选择与用户问题的目的最相关的一个选项"
        )
        res = stream_to_string(self.llm.chat_for_stream(_prompt))
        return remove_think(res)



#
#  Copyright 2024 The InfiniFlow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import json
import logging
import os
import re

from nltk.corpus import wordnet

_curr_dir = os.path.dirname(os.path.abspath(__file__))


class Synonym:
    def __init__(self):
        self.dictionary = None
        path = _curr_dir + "/dictionary/synonym.json"
        with open(path, "r") as f:
            try:
                self.dictionary = json.load(f)
                self.dictionary = {
                    (k.lower() if isinstance(k, str) else k): v
                    for k, v in self.dictionary.items()
                }
            except Exception:
                logging.warning("Missing synonym.json")
                self.dictionary = {}

        if not len(self.dictionary.keys()):
            logging.warning("Fail to load synonym")

    def lookup(self, tk, topn=8):
        if not tk or not isinstance(tk, str):
            return []

        # Check the custom dictionary (both keys and tk are already lowercase)
        key = re.sub(r"[ \t]+", " ", tk.strip())
        res = self.dictionary.get(key, [])
        if isinstance(res, str):
            res = [res]
        if res:  # Found in dictionary → return directly
            return res[:topn]

        # If not found and tk is purely alphabetical → fallback to WordNet
        if re.fullmatch(r"[a-z]+", tk):
            wn_set = {
                re.sub("_", " ", syn.name().split(".")[0])
                for syn in wordnet.synsets(tk)
            }
            wn_set.discard(tk)  # Remove the original token itself
            wn_res = [t for t in wn_set if t]
            return wn_res[:topn]

        # Nothing found in either source
        return []

    def add_syn(self, word: str, synonyms: str) -> None:
        """
        Add a synonym for a word to the dictionary.

        Args:
            word: The word to add synonyms for
            synonyms: A single synonym string
        """
        if not word or not isinstance(word, str):
            logging.warning("Invalid word provided to add_syn")
            return

        if not synonyms:
            logging.warning("No synonyms provided to add_syn")
            return

        # Convert to lowercase for consistency
        word = word.lower()

        # Get existing synonyms for the word
        existing_syns = self.dictionary.get(word, [])

        # Convert existing synonyms to list if it's a string
        if isinstance(existing_syns, str):
            existing_syns = [existing_syns]

        # Convert new synonym to lowercase
        new_synonym = synonyms.lower()

        # Create a set of existing synonyms for quick lookup
        existing_set = {s.lower() for s in existing_syns}

        # Add new synonym if it doesn't already exist
        if new_synonym not in existing_set:
            existing_syns.append(new_synonym)

        # Update the dictionary
        self.dictionary[word] = existing_syns

        logging.info(f"Added synonym '{synonyms}' for '{word}'")


if __name__ == "__main__":
    dl = Synonym()
    print(dl.dictionary)


import logging
import os
import threading

import jsonlines

from ..base import BaseKVStorage
from ..utils import create_file_dir


class JsonKVStorage(BaseKVStorage):

    def __init__(self, namespace: str, working_dir: str = "./duowen_graph_data"):
        super().__init__(namespace=namespace)
        self._file_name = os.path.join(working_dir, f"kv_store_{self.namespace}.jsonl")
        create_file_dir(self._file_name)
        self._data = {}
        if os.path.exists(self._file_name):
            with jsonlines.open(self._file_name, "r") as reader:
                for line in reader:
                    self._data[line["key"]] = line["value"]
        logging.info(f"Load KV {self.namespace} with {len(self._data)} data")
        self._buffer = []
        self._lock = threading.Lock()

    def all_keys(self) -> list[str]:
        return list(self._data.keys())

    def index_done_callback(self):
        with self._lock:
            with jsonlines.open(self._file_name, "a") as writer:
                for i in self._buffer:
                    if i["key"] in self._data:
                        writer.write(i)
                self._buffer = []

    def get_by_id(self, id):
        return self._data.get(id, None)

    def get_by_ids(self, ids, fields=None):
        if fields is None:
            return [self._data.get(id, None) for id in ids]
        return [
            (
                {k: v for k, v in self._data[id].items() if k in fields}
                if self._data.get(id, None)
                else None
            )
            for id in ids
        ]

    def filter_keys(self, data: list[str]) -> set[str]:
        return set([s for s in data if s not in self._data])

    def upsert(self, data: dict[str, dict]):
        with self._lock:
            self._data.update(data)
            for k, v in data.items():
                self._buffer.append({"value": v, "key": k})

    def drop(self):
        with self._lock:
            self._data = {}

    def delete_by_ids(self, ids: list[str]):
        with self._lock:
            for id in ids:
                self._data.pop(id, None)

    def query_done_callback(self):
        pass

    def index_start_callback(self):
        pass



import networkx as nx

from duowen_agent.llm import OpenAIChat
from .base import BaseLLM, GraphChange, BaseVectorStorage, BaseKVStorage
from .prompt import GRAPH_FIELD_SEP
from .utils import (
    print_call_back,
    is_interrupt,
    load_json_graph,
    tidy_graph,
    dump_graph,
    get_from_to,
    NetworkXUtils,
)


class MergeGraph(BaseLLM):
    def __init__(
        self,
        graph_name: str,
        llm_instance: OpenAIChat,
        knowledge_graph_inst: BaseKVStorage,
        entity_vdb: BaseVectorStorage,
        call_back_func: callable = print_call_back,
        interrupt_func: callable = is_interrupt,
        retry_cnt: int = 3,
        retry_sleep: int = 1,
        concurrent_num: int = 1,
        llm_cache: BaseKVStorage = None,
    ):
        super().__init__(
            llm_instance,
            call_back_func,
            interrupt_func,
            retry_cnt,
            retry_sleep,
            concurrent_num,
            llm_cache,
        )
        self.knowledge_graph_inst = knowledge_graph_inst
        self.entity_vdb = entity_vdb
        self.graph_name = graph_name

    def merge(self, subgraph: nx.Graph) -> tuple[nx.Graph, GraphChange]:
        return self.merge_subgraph(subgraph)

    def merge_subgraph(
        self,
        subgraph: nx.Graph,
    ) -> tuple[nx.Graph, GraphChange]:

        change = GraphChange()
        old_graph = self.knowledge_graph_inst.get_by_id(self.graph_name)
        if old_graph is not None:
            old_graph = load_json_graph(old_graph)
            self._call_back_func("与现有图形进行合并")
            tidy_graph(old_graph, self._call_back_func)
            new_graph = self.graph_merge(old_graph, subgraph, change)
        else:
            new_graph = subgraph
            change.added_updated_nodes = set(new_graph.nodes())
            change.added_updated_edges = set(new_graph.edges())

        new_graph = NetworkXUtils(new_graph).pagerank()

        _entity_data = {}
        for i in change.added_updated_nodes:
            _node_data = new_graph.nodes.get(i)
            _entity_data[_node_data["entity_name"]] = (
                f'{_node_data["entity_name"]}: {_node_data["description"]}',
                _node_data,
            )
        self.entity_vdb.upsert(_entity_data)
        self._call_back_func("完成子图合并到全局图")
        self.knowledge_graph_inst.upsert({self.graph_name: dump_graph(new_graph)})
        return new_graph, change

    @staticmethod
    def graph_merge(g1: nx.Graph, g2: nx.Graph, change: GraphChange):
        """Merge graph g2 into g1 in place."""
        for node_name, attr in g2.nodes(data=True):
            change.added_updated_nodes.add(node_name)
            if not g1.has_node(node_name):
                g1.add_node(node_name, **attr)
                continue
            node = g1.nodes[node_name]
            node["description"] += GRAPH_FIELD_SEP + attr["description"]
            # A node's source_id indicates which chunks it came from.
            node["source_id"] += attr["source_id"]

        for source, target, attr in g2.edges(data=True):
            change.added_updated_edges.add(get_from_to(source, target))
            edge = g1.get_edge_data(source, target)
            if edge is None:
                g1.add_edge(source, target, **attr)
                continue
            edge["weight"] += attr.get("weight", 0)
            edge["description"] += GRAPH_FIELD_SEP + attr["description"]
            # edge["keywords"] += attr["keywords"]
            # A edge's source_id indicates which chunks it came from.
            edge["source_id"] += attr["source_id"]

        for node_degree in g1.degree:
            g1.nodes[str(node_degree[0])]["rank"] = int(node_degree[1])
        # A graph's source_id indicates which documents it came from.
        if "source_id" not in g1.graph:
            g1.graph["source_id"] = []
        g1.graph["source_id"] += g2.graph.get("source_id", [])
        return g1



import logging
import queue
from collections.abc import Callable
from concurrent.futures import Future, ThreadPoolExecutor, TimeoutError
from contextlib import ExitStack
from datetime import timedelta
from types import TracebackType
from typing import Any, Generic, Self, TypeVar

from httpx import HTTPStatusError
from pydantic import BaseModel

from duowen_agent.mcp.error import MCPAuthError, MCPConnectionError
from duowen_agent.mcp.types import (
    CancelledNotification,
    ClientNotification,
    ClientRequest,
    ClientResult,
    ErrorData,
    JSONRPCError,
    JSONRPCMessage,
    JSONRPCNotification,
    JSONRPCRequest,
    JSONRPCResponse,
    MessageMetadata,
    RequestId,
    RequestParams,
    ServerMessageMetadata,
    ServerNotification,
    ServerRequest,
    ServerResult,
    SessionMessage,
)

SendRequestT = TypeVar("SendRequestT", ClientRequest, ServerRequest)
SendResultT = TypeVar("SendResultT", ClientResult, ServerResult)
SendNotificationT = TypeVar("SendNotificationT", ClientNotification, ServerNotification)
ReceiveRequestT = TypeVar("ReceiveRequestT", ClientRequest, ServerRequest)
ReceiveResultT = TypeVar("ReceiveResultT", bound=BaseModel)
ReceiveNotificationT = TypeVar(
    "ReceiveNotificationT", ClientNotification, ServerNotification
)
DEFAULT_RESPONSE_READ_TIMEOUT = 1.0


class RequestResponder(Generic[ReceiveRequestT, SendResultT]):
    """Handles responding to MCP requests and manages request lifecycle.

    This class MUST be used as a context manager to ensure proper cleanup and
    cancellation handling:

    Example:
        with request_responder as resp:
            resp.respond(result)

    The context manager ensures:
    1. Proper cancellation scope setup and cleanup
    2. Request completion tracking
    3. Cleanup of in-flight requests
    """

    request: ReceiveRequestT
    _session: Any
    _on_complete: Callable[["RequestResponder[ReceiveRequestT, SendResultT]"], Any]

    def __init__(
        self,
        request_id: RequestId,
        request_meta: RequestParams.Meta | None,
        request: ReceiveRequestT,
        session: """BaseSession[
            SendRequestT,
            SendNotificationT,
            SendResultT,
            ReceiveRequestT,
            ReceiveNotificationT
        ]""",
        on_complete: Callable[["RequestResponder[ReceiveRequestT, SendResultT]"], Any],
    ) -> None:
        self.request_id = request_id
        self.request_meta = request_meta
        self.request = request
        self._session = session
        self._completed = False
        self._on_complete = on_complete
        self._entered = False  # Track if we're in a context manager

    def __enter__(self) -> "RequestResponder[ReceiveRequestT, SendResultT]":
        """Enter the context manager, enabling request cancellation tracking."""
        self._entered = True
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the context manager, performing cleanup and notifying completion."""
        try:
            if self._completed:
                self._on_complete(self)
        finally:
            self._entered = False

    def respond(self, response: SendResultT | ErrorData) -> None:
        """Send a response for this request.

        Must be called within a context manager block.
        Raises:
            RuntimeError: If not used within a context manager
            AssertionError: If request was already responded to
        """
        if not self._entered:
            raise RuntimeError("RequestResponder must be used as a context manager")
        assert not self._completed, "Request already responded to"

        self._completed = True

        self._session._send_response(request_id=self.request_id, response=response)

    def cancel(self) -> None:
        """Cancel this request and mark it as completed."""
        if not self._entered:
            raise RuntimeError("RequestResponder must be used as a context manager")

        self._completed = True  # Mark as completed so it's removed from in_flight
        # Send an error response to indicate cancellation
        self._session._send_response(
            request_id=self.request_id,
            response=ErrorData(code=0, message="Request cancelled", data=None),
        )


class BaseSession(
    Generic[
        SendRequestT,
        SendNotificationT,
        SendResultT,
        ReceiveRequestT,
        ReceiveNotificationT,
    ],
):
    """
    Implements an MCP "session" on top of read/write streams, including features
    like request/response linking, notifications, and progress.

    This class is a context manager that automatically starts processing
    messages when entered.
    """

    _response_streams: dict[RequestId, queue.Queue[JSONRPCResponse | JSONRPCError]]
    _request_id: int
    _in_flight: dict[RequestId, RequestResponder[ReceiveRequestT, SendResultT]]
    _receive_request_type: type[ReceiveRequestT]
    _receive_notification_type: type[ReceiveNotificationT]

    def __init__(
        self,
        read_stream: queue.Queue,
        write_stream: queue.Queue,
        receive_request_type: type[ReceiveRequestT],
        receive_notification_type: type[ReceiveNotificationT],
        # If none, reading will never time out
        read_timeout_seconds: timedelta | None = None,
    ) -> None:
        self._read_stream = read_stream
        self._write_stream = write_stream
        self._response_streams = {}
        self._request_id = 0
        self._receive_request_type = receive_request_type
        self._receive_notification_type = receive_notification_type
        self._session_read_timeout_seconds = read_timeout_seconds
        self._in_flight = {}
        self._exit_stack = ExitStack()
        # Initialize executor and future to None for proper cleanup checks
        self._executor: ThreadPoolExecutor | None = None
        self._receiver_future: Future | None = None

    def __enter__(self) -> Self:
        # The thread pool is dedicated to running `_receive_loop`. Setting `max_workers` to 1
        # ensures no unnecessary threads are created.
        self._executor = ThreadPoolExecutor(max_workers=1)
        self._receiver_future = self._executor.submit(self._receive_loop)
        return self

    def check_receiver_status(self) -> None:
        """`check_receiver_status` ensures that any exceptions raised during the
        execution of `_receive_loop` are retrieved and propagated."""
        if self._receiver_future and self._receiver_future.done():
            self._receiver_future.result()

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self._read_stream.put(None)
        self._write_stream.put(None)

        # Wait for the receiver loop to finish
        if self._receiver_future:
            try:
                self._receiver_future.result(timeout=5.0)  # Wait up to 5 seconds
            except TimeoutError:
                # If the receiver loop is still running after timeout, we'll force shutdown
                pass

        # Shutdown the executor
        if self._executor:
            self._executor.shutdown(wait=True)

    def send_request(
        self,
        request: SendRequestT,
        result_type: type[ReceiveResultT],
        request_read_timeout_seconds: timedelta | None = None,
        metadata: MessageMetadata = None,
    ) -> ReceiveResultT:
        """
        Sends a request and wait for a response. Raises an McpError if the
        response contains an error. If a request read timeout is provided, it
        will take precedence over the session read timeout.

        Do not use this method to emit notifications! Use send_notification()
        instead.
        """
        self.check_receiver_status()

        request_id = self._request_id
        self._request_id = request_id + 1

        response_queue: queue.Queue[JSONRPCResponse | JSONRPCError] = queue.Queue()
        self._response_streams[request_id] = response_queue

        try:
            jsonrpc_request = JSONRPCRequest(
                jsonrpc="2.0",
                id=request_id,
                **request.model_dump(by_alias=True, mode="json", exclude_none=True),
            )

            self._write_stream.put(
                SessionMessage(
                    message=JSONRPCMessage(jsonrpc_request), metadata=metadata
                )
            )
            timeout = DEFAULT_RESPONSE_READ_TIMEOUT
            if request_read_timeout_seconds is not None:
                timeout = float(request_read_timeout_seconds.total_seconds())
            elif self._session_read_timeout_seconds is not None:
                timeout = float(self._session_read_timeout_seconds.total_seconds())
            while True:
                try:
                    response_or_error = response_queue.get(timeout=timeout)
                    break
                except queue.Empty:
                    self.check_receiver_status()
                    continue

            if response_or_error is None:
                raise MCPConnectionError(
                    ErrorData(
                        code=500,
                        message="No response received",
                    )
                )
            elif isinstance(response_or_error, JSONRPCError):
                if response_or_error.error.code == 401:
                    raise MCPAuthError(
                        ErrorData(
                            code=response_or_error.error.code,
                            message=response_or_error.error.message,
                        )
                    )
                else:
                    raise MCPConnectionError(
                        ErrorData(
                            code=response_or_error.error.code,
                            message=response_or_error.error.message,
                        )
                    )
            else:
                return result_type.model_validate(response_or_error.result)

        finally:
            self._response_streams.pop(request_id, None)

    def send_notification(
        self,
        notification: SendNotificationT,
        related_request_id: RequestId | None = None,
    ) -> None:
        """
        Emits a notification, which is a one-way message that does not expect
        a response.
        """
        self.check_receiver_status()

        # Some transport implementations may need to set the related_request_id
        # to attribute to the notifications to the request that triggered them.
        jsonrpc_notification = JSONRPCNotification(
            jsonrpc="2.0",
            **notification.model_dump(by_alias=True, mode="json", exclude_none=True),
        )
        session_message = SessionMessage(
            message=JSONRPCMessage(jsonrpc_notification),
            metadata=(
                ServerMessageMetadata(related_request_id=related_request_id)
                if related_request_id
                else None
            ),
        )
        self._write_stream.put(session_message)

    def _send_response(
        self, request_id: RequestId, response: SendResultT | ErrorData
    ) -> None:
        if isinstance(response, ErrorData):
            jsonrpc_error = JSONRPCError(jsonrpc="2.0", id=request_id, error=response)
            session_message = SessionMessage(message=JSONRPCMessage(jsonrpc_error))
            self._write_stream.put(session_message)
        else:
            jsonrpc_response = JSONRPCResponse(
                jsonrpc="2.0",
                id=request_id,
                result=response.model_dump(
                    by_alias=True, mode="json", exclude_none=True
                ),
            )
            session_message = SessionMessage(message=JSONRPCMessage(jsonrpc_response))
            self._write_stream.put(session_message)

    def _receive_loop(self) -> None:
        """
        Main message processing loop.
        In a real synchronous implementation, this would likely run in a separate thread.
        """
        while True:
            try:
                # Attempt to receive a message (this would be blocking in a synchronous context)
                message = self._read_stream.get(timeout=DEFAULT_RESPONSE_READ_TIMEOUT)
                if message is None:
                    break
                if isinstance(message, HTTPStatusError):
                    response_queue = self._response_streams.get(self._request_id - 1)
                    if response_queue is not None:
                        response_queue.put(
                            JSONRPCError(
                                jsonrpc="2.0",
                                id=self._request_id - 1,
                                error=ErrorData(
                                    code=message.response.status_code,
                                    message=message.args[0],
                                ),
                            )
                        )
                    else:
                        self._handle_incoming(
                            RuntimeError(
                                f"Received response with an unknown request ID: {message}"
                            )
                        )
                elif isinstance(message, Exception):
                    self._handle_incoming(message)
                elif isinstance(message.message.root, JSONRPCRequest):
                    validated_request = self._receive_request_type.model_validate(
                        message.message.root.model_dump(
                            by_alias=True, mode="json", exclude_none=True
                        )
                    )

                    responder = RequestResponder(
                        request_id=message.message.root.id,
                        request_meta=(
                            validated_request.root.params.meta
                            if validated_request.root.params
                            else None
                        ),
                        request=validated_request,
                        session=self,
                        on_complete=lambda r: self._in_flight.pop(r.request_id, None),
                    )

                    self._in_flight[responder.request_id] = responder
                    self._received_request(responder)

                    if not responder._completed:
                        self._handle_incoming(responder)

                elif isinstance(message.message.root, JSONRPCNotification):
                    try:
                        notification = self._receive_notification_type.model_validate(
                            message.message.root.model_dump(
                                by_alias=True, mode="json", exclude_none=True
                            )
                        )
                        # Handle cancellation notifications
                        if isinstance(notification.root, CancelledNotification):
                            cancelled_id = notification.root.params.requestId
                            if cancelled_id in self._in_flight:
                                self._in_flight[cancelled_id].cancel()
                        else:
                            self._received_notification(notification)
                            self._handle_incoming(notification)
                    except Exception as e:
                        # For other validation errors, log and continue
                        logging.warning(
                            "Failed to validate notification: %s. Message was: %s",
                            e,
                            message.message.root,
                        )
                else:  # Response or error
                    response_queue = self._response_streams.get(message.message.root.id)
                    if response_queue is not None:
                        response_queue.put(message.message.root)
                    else:
                        self._handle_incoming(RuntimeError(f"Server Error: {message}"))
            except queue.Empty:
                continue
            except Exception as e:
                logging.exception("Error in message processing loop")
                raise

    def _received_request(
        self, responder: RequestResponder[ReceiveRequestT, SendResultT]
    ) -> None:
        """
        Can be overridden by subclasses to handle a request without needing to
        listen on the message stream.

        If the request is responded to within this method, it will not be
        forwarded on to the message stream.
        """
        pass

    def _received_notification(self, notification: ReceiveNotificationT) -> None:
        """
        Can be overridden by subclasses to handle a notification without needing
        to listen on the message stream.
        """
        pass

    def send_progress_notification(
        self, progress_token: str | int, progress: float, total: float | None = None
    ) -> None:
        """
        Sends a progress notification for a request that is currently being
        processed.
        """
        pass

    def _handle_incoming(
        self,
        req: (
            RequestResponder[ReceiveRequestT, SendResultT]
            | ReceiveNotificationT
            | Exception
        ),
    ) -> None:
        """A generic handler for incoming messages. Overwritten by subclasses."""
        pass
